from ftw.upgrade import UpgradeStep
from ftw.shop.interfaces import IVariationConfig


class ConvertVariationsToMappingInAnnotationStorage(UpgradeStep):
    """Convert variations to mapping in annotation storage.
    """

    def __call__(self):
        self.install_upgrade_profile()

        for item in self.objects({'portal_type': 'ShopItem'},
                                 'Migrate variations storage'):

            varConf = IVariationConfig(item)
            item_data = varConf.getVariationDict()
            varConf.updateVariationConfig(item_data)



from ftw.shop.content.shopcategory import ShopCategory
from ftw.upgrade import UpgradeStep
from plone.uuid.interfaces import IUUID


class ConvertCategoryRankKeysToUid(UpgradeStep):
    """Convert category rank keys to uid.
    """

    def __call__(self):
        self.install_upgrade_profile()
        objs = self.objects(
            {
                'object_provides': [
                    'ftw.shop.interfaces.IShopCategory',
                    'ftw.shop.interfaces.IShopItem',
                ]
            },
            message='Converting category rank keys to uid',
        )
        for obj in objs:
            if hasattr(obj, '_categoryRanks'):
                for category, rank in obj._categoryRanks.items()[:]:
                    if isinstance(category, ShopCategory):
                        # Only convert the keys the first time the upgrade
                        # step is run.
                        obj._categoryRanks[IUUID(category)] = rank
                        del obj._categoryRanks[category]



from Acquisition import aq_parent
from ftw.shop import shopMessageFactory as _
from ftw.shop.interfaces import IShopCategory
from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from Products.statusmessages.interfaces import IStatusMessage


class ManageCategories(BrowserView):

    template = ViewPageTemplateFile('templates/edit_categories.pt')

    def __call__(self):
        """
        Self-submitting form that lets the user edit categories that
        ShopItems or ShopCategories belong to
        """
        form = self.request.form

        # Make sure we had a proper form submit, not just a GET request
        submitted = form.get('form.submitted', False)
        if submitted:
            edited_category_data = self.request.get('categories', [])
            self.update_categories(edited_category_data)
            IStatusMessage(self.request).addStatusMessage(
                _(u'msg_categories_updated',
                  default=u"Categories updated."), type="info")
            self.request.RESPONSE.redirect(self.context.absolute_url())

        return self.template()

    def update_categories(self, categories):
        """Update the categories a ShopItem or ShopCategory belongs to
        """
        #categories = self.request.get('categories', None)
        old_uids = [obj.UID() for obj in self.context.listCategories()]

        # add the checked categories
        new_uids = [uid for uid in categories if uid not in old_uids]
        for c in new_uids:
            self.context.addToCategory(c)

        # remove the unchecked categories
        uids_delete = [uid for uid in old_uids if uid not in categories]
        categories_delete = uids_delete
        for c in categories_delete:
            self.context.removeFromCategory(c)

        #update ranks
        for c in self.context.listCategories():
            rank = self.context.REQUEST.get('rank_%s' % c.UID(), 1)
            self.context.setRankForCategory(c, rank)

        putils = getToolByName(self.context, 'plone_utils')
        putils.addPortalMessage(_(u'msg_categories_updated',
                                  default=u"Categories updated."), 'info')
        return self.context.REQUEST.RESPONSE.redirect('%s' %
                                            (self.context.absolute_url()))

    def find_first_non_category(self):
        """Walk up from the current context and find the first object that
        isn't of type ShopCategory.
        """
        # Start with the ShopItem's parent (which is a ShopCategory)
        parent = aq_parent(self.context)
        while IShopCategory.providedBy(parent):
            parent = aq_parent(parent)
        return parent

    def get_top_level_categories(self):
        """Walking up from current context, find the top level categories
        defining the start of a shop category sub-structure.
        """
        category_tree_container = self.find_first_non_category()

        catalog = getToolByName(self.context, 'portal_catalog')
        query = dict(
            object_provides=IShopCategory.__identifier__,
            path = {
                'query': '/'.join(category_tree_container.getPhysicalPath()),
                'depth': 1},
            sort_on = 'path',
            )
        categories = [b.getObject() for b in catalog(query)]
        return categories

    def get_sub_categories(self, context):
        """Given a ShopCategory as `context`, return all its subcategories.
        """
        catalog = getToolByName(self.context, 'portal_catalog')
        query = dict(
            object_provides=IShopCategory.__identifier__,
            path = {
                'query': '/'.join(context.getPhysicalPath()),
                },
            sort_on = 'path',
            )
        categories = [b.getObject() for b in catalog(query)]
        # Don't include the queried context itself
        categories.remove(context)
        return categories

    def list_all_categories(self, categoryUID):
        """return all Category instances except the context
        (that could be a Category instance too).
        """
        categoryUID = self.request.get('CategoryUID', None)
        #portalPath = getToolByName(self.context, 'portal_url').getPortalPath()
        #ltool = getToolByName(self.context, 'portal_languages')
        catalog = getToolByName(self.context, 'portal_catalog')
        #lang = ltool.getPreferredLanguage()

        query = {}
        query['portal_type'] = ['ShopCategory']
        #query['path'] = '%s/%s' % (portalPath, lang)
        query['sort_on'] = 'fullTitle'

        categories = catalog(query)
        return [category.getObject() for category in categories
                                        if category.UID != categoryUID]



from ftw.builder import Builder
from ftw.builder import create
from ftw.shop.browser.cart import ShoppingCartAdapter
from ftw.shop.testing import FTW_SHOP_INTEGRATION_TESTING
from plone.app.testing import login
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.uuid.interfaces import IUUID
from unittest2 import TestCase


class TestBrowserCart(TestCase):

    layer = FTW_SHOP_INTEGRATION_TESTING

    def setUp(self):
        portal = self.layer['portal']
        portal.REQUEST.set('SESSION', {})
        setRoles(portal, TEST_USER_ID, ['Manager'])
        login(portal, TEST_USER_NAME)

        self.supplier = create(Builder('supplier')
                               .titled('Demo Supplier')
                               .having(email='demo@email.com'))
        self.category = create(Builder('shop category').titled('Category'))
        self.subcategory = create(Builder('shop category')
                                  .within(self.category)
                                  .titled('SubCategory'))
        self.shopitem = create(Builder('shop item').within(self.subcategory))

    def test_get_supplier_from_shopitem(self):
        cart = ShoppingCartAdapter(self.shopitem, self.shopitem.REQUEST)
        cart.add_item()

        self.shopitem.setSupplier(IUUID(self.supplier))
        self.assertEquals(self.supplier, cart._get_supplier(self.shopitem))

    def test_get_supplier_from_parent(self):
        cart = ShoppingCartAdapter(self.shopitem, self.shopitem.REQUEST)
        cart.add_item()

        self.subcategory.setSupplier(IUUID(self.supplier))
        self.assertEquals(self.supplier, cart._get_supplier(self.shopitem))

    def test_get_supplier_recursively(self):
        cart = ShoppingCartAdapter(self.shopitem, self.shopitem.REQUEST)
        cart.add_item()

        self.category.setSupplier(IUUID(self.supplier))
        self.assertEquals(self.supplier, cart._get_supplier(self.shopitem))



# -*- coding: utf-8 -*-
"""Definition of the Shop Item content type
"""

from Acquisition import aq_parent
from Products.Archetypes import atapi
from Products.ATContentTypes.content.base import ATCTContent
from Products.ATContentTypes.content.schemata import ATContentTypeSchema
from Products.ATContentTypes.configuration import zconf
from Products.Archetypes.public import AnnotationStorage

from zope.interface import implements, alsoProvides

try:
    # Plone 4
    from zope.lifecycleevent import ObjectRemovedEvent
except ImportError:
    # Plone 3
    from zope.app.container.contained import ObjectRemovedEvent


from Products.ATContentTypes.config import HAS_LINGUA_PLONE
if HAS_LINGUA_PLONE:
    from Products.LinguaPlone.public import registerType
else:
    from Products.Archetypes.atapi import registerType

from ftw.shop.interfaces import IBuyable
from ftw.shop.interfaces import IShopItem
from ftw.shop.interfaces import IVariationConfig
from ftw.shop.content.categorizeable import Categorizeable
from ftw.shop.config import PROJECTNAME
from ftw.shop import shopMessageFactory as _


selectable_dimensions = {
    'no_dimensions': {
        'label': _(u"label_no_dimensions", default=u"---"),
        'dimension_unit': None,
        'dimensions': []
    },
    'length_mm_mm': {
        'label': _(u"label_length", default=u"Length"),
        'dimension_unit': u'mm',
        'dimensions': [_(u"Length (mm)")]
    },
    'length_m_m': {
        'label': _(u"label_length", default=u"Length"),
        'dimension_unit': u'm',
        'dimensions': [_(u"Length (m)")]
    },
    'length_width_mm_mm2': {
        'label': _(u"label_l_w", default=u"Length, Width"),
        'dimension_unit': u'mm²',
        'dimensions': [
            _(u"Length (mm)"),
            _(u"Width (mm)")]
    },
    'length_width_mm_m2': {
        'label': _(u"label_l_w", default=u"Length, Width"),
        'price_unit': u'm²',
        'price_to_dimension_modifier': 1000000,
        'dimension_unit': u'mm²',
        'dimensions': [
            _(u"Length (mm)"),
            _(u"Width (mm)")]
    },
    'length_width_thickness_mm_mm3': {
        'label': _(u"label_l_w_t", default=u"Length, Width, Thickness"),
        'dimension_unit': u'mm³',
        'dimensions': [
            _(u"Length (mm)"),
            _(u"Width (mm)"),
            _(u"Thickness (mm)")]
    },
    'length_width_thickness_mm_m3': {
        'label': _(u"label_l_w_t", default=u"Length, Width, Thickness"),
        'price_unit': u'm³',
        'price_to_dimension_modifier': 1000000000,
        'dimension_unit': u'mm³',
        'dimensions': [
            _(u"Length (mm)"),
            _(u"Width (mm)"),
            _(u"Thickness (mm)")]
    },
    'weight_g_g': {
        'label': _(u'label_weight', default=u"Weight"),
        'dimension_unit': u'g',
        'dimensions': [_(u"Weight (g)")]
    },
    'weight_g_kg': {
        'label': _(u'label_weight', default=u"Weight"),
        'price_unit': u'kg',
        'price_to_dimension_modifier': 1000,
        'dimension_unit': u'g',
        'dimensions': [_(u"Weight (g)")]
    },
    'weight_kg_kg': {
        'label': _(u'label_weight', default=u"Weight"),
        'dimension_unit': u'kg',
        'dimensions': [_(u"Weight (kg)")]
    }
}

ShopItemSchema = ATContentTypeSchema.copy() + atapi.Schema((

        atapi.ImageField(
            'image',
            required=False,
            languageIndependent=True,
            widget = atapi.ImageWidget(
                label = _(u"label_image", default=u"Image"),
            ),
            storage=AnnotationStorage(),
                sizes= {'large': (768, 768),
                         'mini': (200, 200),
                         'thumb': (128, 128),
                  },
            ),


        atapi.TextField(
            'text',
            required=False,
            searchable=True,
            primary=True,
            storage=atapi.AnnotationStorage(migrate=True),
            allowable_content_types=('text/html', ),
            default_content_type='text/html',
            validators=('isTidyHtmlWithCleanup', ),
            default_input_type='text/html',
            default_output_type='text/x-html-safe',
            widget=atapi.RichWidget(
                description='',
                label=_(u'label_body_text', default=u'Body Text'),
                rows=25,
                allow_file_upload = zconf.ATDocument.allow_document_upload),
            ),
        atapi.FixedPointField(
            'price',
            default = "0.00",
            required = 0,
            languageIndependent=True,
            widget = atapi.DecimalWidget(
                label = _(u"label_price", default=u"Price"),
                description = _(u"desc_price", default=u""),
                size=8,
                ),
            ),

        atapi.BooleanField(
            'showPrice',
            default = False,
            languageIndependent=True,
            widget = atapi.BooleanWidget(
                label = _(u"label_show_price", default=u"Show price"),
                description = _(u"desc_show_price", default=u""),
                ),
            ),

        atapi.StringField(
            'skuCode',
            required = 0,
            languageIndependent=True,
            widget = atapi.StringWidget(
                label = _(u"label_sku_code", default=u"SKU code"),
                description = _(u"desc_sku_code", default=u""),
                ),
            ),

        atapi.TextField(
            name='unit',
            searchable=True,
            widget=atapi.StringWidget(
                label=_(u'label_unit', default='Unit'),
                description=_(
                    u'description_unit',
                    default='The unit for the product quantity.'),
            ),
            schemata='default',
        ),

        atapi.StringField(
            'selectable_dimensions',
            required=True,
            vocabulary_factory='ftw.shop.selectable_dimensions_vocabulary',
            widget=atapi.SelectionWidget(
                format='select',
                label=_(u'label_selectable_dimensions',
                        default=u'Selectable dimensions'),
                description=_(
                    u'description_dimensions',
                    default='Specifies which dimensions of the product can '
                            'be chosen by the user. The price is per base unit '
                            '(mm, g, etc.).')
            ),
            schemata='default'
        ),

        atapi.StringField(
            'variation1_attribute',
            required = 0,
            widget = atapi.StringWidget(
                label = _(u"label_variation1_attr",
                          default=u"Variation 1 Attribute"),
                description = _(u"desc_variation1_attr", default=u""),
                visible={'view': 'invisible', 'edit': 'invisible'},
                ),
            ),

        atapi.LinesField(
            'variation1_values',
            required = 0,
            widget = atapi.LinesWidget(
                label = _(u"label_variation1_values",
                          default=u"Variation 1 Values"),
                description = _(u"desc_variation1_values", default=u""),
                visible={'view': 'invisible', 'edit': 'invisible'},
                ),
            ),


        atapi.StringField(
            'variation2_attribute',
            required = 0,
            widget = atapi.StringWidget(
                label = _(u"label_variation2_attr",
                          default=u"Variation 2 Attribute"),
                description = _(u"desc_variation2_attr", default=u""),
                visible={'view': 'invisible', 'edit': 'invisible'},
                ),
            ),

        atapi.LinesField(
            'variation2_values',
            required = 0,
            widget = atapi.LinesWidget(
                label = _(u"label_variation2_values",
                          default=u"Variation 2 Values"),
                description = _(u"desc_variation2_values", default=u""),
                visible={'view': 'invisible', 'edit': 'invisible'},
                ),
            ),

        atapi.ReferenceField(
            'supplier',
            required = 0,
            languageIndependent=True,
            relationship = 'item_supplier',
            vocabulary_factory="ftw.shop.suppliers_vocabulary",
            widget = atapi.ReferenceWidget(
                label = _(u"label_supplier", default=u"Supplier"),
                description = _(u"desc_supplier", default=u""),
                checkbox_bound = 10,
                ),
            ),
        ))


class ShopItem(Categorizeable, ATCTContent):
    """A simple shop item"""
    implements(IShopItem)
    alsoProvides(IBuyable)

    meta_type = "ShopItem"
    schema = ShopItemSchema

    def SearchableText(self):
        """ Make variations searchable. """
        data = super(ShopItem, self).SearchableText()

        var_conf = IVariationConfig(self)
        variation_details = []
        if var_conf.hasVariations():
            for variation in var_conf.getVariationDict().values():
                if not variation['active']:
                    continue
                variation_details.append(variation['description'])
                variation_details.append(variation['skuCode'])

        return ' '.join([
            data,
            ' '.join(self.getVariation1_values()),
            ' '.join(self.getVariation2_values()),
            ' '.join(variation_details),
            self.getSkuCode()
        ])

    def getDimensionDict(self):
        dim_key = self.Schema().getField('selectable_dimensions').get(self)
        if not dim_key or dim_key not in selectable_dimensions:
            return selectable_dimensions['no_dimensions']

        return selectable_dimensions[dim_key]

    def getSelectableDimensions(self):
        dimension_dict = self.getDimensionDict()
        return dimension_dict.get('dimensions', [])

    def getDimensionsLabel(self):
        dimension_dict = self.getDimensionDict()
        return dimension_dict.get('dimension_unit', None)

    def getPriceUnit(self):
        dimension_dict = self.getDimensionDict()
        if 'price_unit' in dimension_dict:
            return dimension_dict['price_unit']
        return dimension_dict.get('dimension_unit', None)

    def getPriceModifier(self):
        dimension_dict = self.getDimensionDict()
        return dimension_dict.get('price_to_dimension_modifier', 1)


def add_to_containing_category(context, event):
    """
    @param context: Zope object for which the event was fired for.
    Usually this is Plone content object.

    @param event: Subclass of event.
    """

    if not event.__class__ ==  ObjectRemovedEvent:
        parent = aq_parent(context)
        context.addToCategory(parent.UID())

registerType(ShopItem, PROJECTNAME)



from ftw.builder import Builder
from ftw.builder import create
from ftw.shop.interfaces import IShopConfiguration
from ftw.shop.interfaces import IShopRoot
from ftw.shop.testing import FTW_SHOP_FUNCTIONAL_TESTING
from ftw.shop.tests.helpers import get_mail_header
from ftw.shop.tests.pages import checkout
from ftw.testbrowser import browsing
from ftw.testing.mailing import Mailing
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import login
from plone.app.testing import setRoles
from plone.registry.interfaces import IRegistry
from unittest2 import TestCase
from zope.component import getUtility
from zope.interface import alsoProvides
import email
import email.header
import transaction


class TestCheckoutMailToCustomer(TestCase):

    layer = FTW_SHOP_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        setRoles(self.portal, TEST_USER_ID, ['Manager'])
        login(self.portal, TEST_USER_NAME)

        alsoProvides(self.portal, IShopRoot)
        create(Builder('cart portlet'))
        Mailing(self.portal).set_up()

    def checkout_and_get_mail(self, *items):
        mailing = Mailing(self.portal)
        for item in items:
            checkout.visit_checkout_with_one_item_in_cart(item)
        checkout.goto(checkout.ORDER_REVIEW).finish()
        self.assertEquals(1, len(mailing.get_messages()),
                          'Notifying shop owner is disabled by default, therefore'
                          ' only one mail (to the customer) should be sent.')
        return email.message_from_string(mailing.pop())

    def open_mail_in_browser(self, mail, browser):
        browser.open_html(mail.get_payload(decode=True))

    @browsing
    def test_customer_is_notified(self, browser):
        mail = self.checkout_and_get_mail()
        self.assertEquals(['Hugo Boss <hugo@boss.com>'], get_mail_header(mail, 'To'))
        self.assertEquals(['Your Webshop Order'], get_mail_header(mail, 'Subject'))

    @browsing
    def test_order_details(self, browser):
        self.open_mail_in_browser(self.checkout_and_get_mail(), browser)
        self.assertRegexpMatches(browser.css('#order-details').first.text,
                                 r'^Your order number is: \d*$')

    @browsing
    def test_personal_information(self, browser):
        self.open_mail_in_browser(self.checkout_and_get_mail(), browser)

        self.assertMultiLineEqual(
            '\n'.join(('Sir',
                       'Hugo Boss',
                       'Example Street 15',
                       '3000 Bern',
                       'Switzerland',
                       '',
                       'Email: hugo@boss.com',
                       'Phone: 001 0101 0101 01')),

            browser.css('#personal-information').first.text)

    @browsing
    def test_shipping_address(self, browser):
        self.open_mail_in_browser(self.checkout_and_get_mail(), browser)

        self.assertMultiLineEqual(
            '\n'.join(('Sir',
                       'Hugo Boss',
                       'Example Street 15',
                       '3000 Bern')),

            browser.css('#shipping-address').first.text)

    @browsing
    def test_product_is_listed_in_order_table(self, browser):
        item = create(Builder('shop item')
                      .titled('A pair of socks')
                      .having(skuCode='123.1',
                              price='12.90',
                              showPrice=True))

        self.open_mail_in_browser(self.checkout_and_get_mail(item), browser)

        self.assertEquals([{'Article number': '123.1',
                            'Product': 'A pair of socks',
                            'Price per item': '12.90',
                            'Quantity': '1',
                            'Total': '12.90'}],
                          browser.css('table').first.dicts(foot=False))

    @browsing
    def test_total_price_is_correct_in_order_table(self, browser):
        items = [create(Builder('shop item').having(price='10.00')),
                 create(Builder('shop item').having(price='5.00'))]

        self.open_mail_in_browser(self.checkout_and_get_mail(*items), browser)

        self.assertEquals([['VAT', '', '', '', '0.00'],
                           ['Total (incl.VAT)', '', '', '', '15.00']],
                          browser.css('table').first.lists(head=False, body=False))

    @browsing
    def test_default_contact_information(self, browser):
        self.open_mail_in_browser(self.checkout_and_get_mail(), browser)

        self.assertEquals(
            'For inquiries please contact us: E-Mail webshop@example.org',

            browser.css('#contact-information').first.text)

    @browsing
    def test_phone_in_contact_information_when_configured(self, browser):
        registry = getUtility(IRegistry)
        settings = registry.forInterface(IShopConfiguration)
        settings.phone_number = u'01 234 56 78'
        transaction.commit()

        self.open_mail_in_browser(self.checkout_and_get_mail(), browser)

        self.assertEquals(
            'For inquiries please contact us: Phone 01 234 56 78,'
            ' E-Mail webshop@example.org',

            browser.css('#contact-information').first.text)



from ftw.builder import Builder
from ftw.builder import create
from ftw.shop.testing import FTW_SHOP_FUNCTIONAL_TESTING
from ftw.shop.tests.pages import cartportlet
from ftw.shop.tests.pages import shopitem
from ftw.testbrowser import browsing
from ftw.testbrowser.pages import plone
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import login
from plone.app.testing import setRoles
from unittest2 import TestCase
import transaction


class TestBrowserCart(TestCase):

    layer = FTW_SHOP_FUNCTIONAL_TESTING

    def setUp(self):
        portal = self.layer['portal']
        setRoles(portal, TEST_USER_ID, ['Manager'])
        login(portal, TEST_USER_NAME)

        create(Builder('cart portlet'))
        transaction.commit()

    @browsing
    def test_adding_items_to_cart(self, browser):
        browser.login().open()
        self.assertTrue(cartportlet.is_visible(), 'Cart portlet should be visible.')
        self.assertTrue(cartportlet.is_empty(), 'Cart portlet should be empty.')

        item = create(Builder('shop item')
                      .titled('A pair of socks'))
        browser.visit(item)
        shopitem.add_to_cart()
        self.assertEquals(['A pair of socks'], cartportlet.items())

    @browsing
    def test_edit_cart_link_visible_when_cart_has_items(self, browser):
        browser.login().open()
        self.assertFalse(cartportlet.edit_cart_link(),
                         '"Edit cart" link should not be visible yet.')

        item = create(Builder('shop item'))
        browser.visit(item)
        shopitem.add_to_cart()
        self.assertTrue(cartportlet.edit_cart_link(),
                        '"Edit cart" link should now be visible.')

    @browsing
    def test_edit_cart_link_opens_edit_cart_view(self, browser):
        item = create(Builder('shop item'))
        browser.login().visit(item)
        shopitem.add_to_cart()
        cartportlet.edit_cart_link().click()
        self.assertEquals('cart_edit', plone.view())

    @browsing
    def test_order_link_opens_checkout_wizard(self, browser):
        item = create(Builder('shop item'))
        browser.login().visit(item)
        shopitem.add_to_cart()
        cartportlet.order_link().click()
        self.assertEquals('checkout-wizard', plone.view())

    @browsing
    def test_order_manager_link_visible_for_superuser(self, browser):
        browser.login().open()
        self.assertTrue(
            cartportlet.order_manager_link(),
            'The test user (Manager) should see the "Order Manager" link.')

    @browsing
    def test_order_manager_link_invisible_for_normal_user(self, browser):
        normal_user = create(Builder('user'))
        browser.login(normal_user.getId()).open()
        self.assertFalse(
            cartportlet.order_manager_link(),
            'The normal user (Member) should NOT see the "Order Manager" link.')

    @browsing
    def test_order_manager_link_points_to_order_manager_view(self, browser):
        browser.login().open()
        cartportlet.order_manager_link().click()
        self.assertEquals('order_manager', plone.view())



"""Definition of the Shop Category content type
"""

from Acquisition import aq_parent, aq_inner
from Products.Archetypes import atapi
from Products.ATContentTypes.content.folder import ATFolder
from Products.ATContentTypes.content.folder import ATFolderSchema
from zope.interface import implements

try:
    # Plone 4
    from zope.lifecycleevent import ObjectRemovedEvent
except ImportError:
    # Plone 3
    from zope.app.container.contained import ObjectRemovedEvent

from Products.ATContentTypes.config import HAS_LINGUA_PLONE
if HAS_LINGUA_PLONE:
    from Products.LinguaPlone.public import registerType
else:
    from Products.Archetypes.atapi import registerType

from ftw.shop import shopMessageFactory as _
from ftw.shop.interfaces import IShopCategory
from ftw.shop.config import PROJECTNAME
from ftw.shop.content.categorizeable import Categorizeable


ShopCategorySchema = ATFolderSchema.copy() + atapi.Schema((
        atapi.ReferenceField(
            'supplier',
            required = 0,
            languageIndependent=True,
            relationship = 'item_supplier',
            vocabulary_factory="ftw.shop.suppliers_vocabulary",
            widget = atapi.ReferenceWidget(
                label = _(u"label_supplier", default=u"Supplier"),
                description = _(u"desc_supplier", default=u""),
                checkbox_bound = 10,
                ),
            ),
        ))


class ShopCategory(Categorizeable, ATFolder):
    """A category for shop items"""
    implements(IShopCategory)

    meta_type = "ShopCategory"
    schema = ShopCategorySchema

    def fullTitle(self):
        """Returns a fully qualified title for the category, including
        parent categories' titles if applicable.
        """
        parent = self
        title_chain = []
        while parent.portal_type =='ShopCategory':
            title_chain.append(parent.Title())
            parent = aq_parent(aq_inner(parent))
        title_chain.reverse()
        return title_chain

registerType(ShopCategory, PROJECTNAME)


def add_to_containing_category(context, event):
    """
    @param context: Zope object for which the event was fired for.
    Usually this is Plone content object.

    @param event: Subclass of event.
    """
    if not event.__class__ ==  ObjectRemovedEvent:
        parent = aq_parent(context)
        if parent.portal_type == 'ShopCategory':
            context.addToCategory(parent.UID())



import transaction
from decimal import Decimal
from Products.CMFCore.utils import getToolByName
import simplejson

from Acquisition import aq_inner
from Products.Five.browser import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from Products.statusmessages.interfaces import IStatusMessage
from plone.i18n.normalizer.interfaces import IIDNormalizer
from zope.component import getUtility
from zope.component import getMultiAdapter
from zope.i18n import translate

from ftw.shop import shopMessageFactory as _
from ftw.shop.interfaces import IVariationConfig


class ShopItemView(BrowserView):
    """Default view for a shop item
    """

    __call__ = ViewPageTemplateFile('templates/shopitem.pt')

    single_item_template = ViewPageTemplateFile('templates/listing/single_item.pt')
    one_variation_template = ViewPageTemplateFile('templates/listing/one_variation_compact.pt')
    two_variations_template = ViewPageTemplateFile('templates/listing/two_variations_compact.pt')

    @property
    def depth(self):
        return len(self.getVariationAttributes())

    def getItems(self):
        """Returns a list with this item as its only element,
        so the listing viewlet can treat it like a list of items
        """
        context = aq_inner(self.context)
        return [context]

    def single_item(self, item):
        return self.single_item_template(item=item)

    def one_variation(self, item):
        return self.one_variation_template(item=item)

    def two_variations(self, item):
        return self.two_variations_template(item=item)

    def getItemDatas(self):
        """Returns a dictionary of an item's properties to be used in
        templates. If the item has variations, the variation config is
        also included.
        """
        results = []
        for item in self.getItems():
            varConf = IVariationConfig(item)
            has_variations = varConf.hasVariations()

            image = item.getField('image')
            if image and image.get_size(item):
                hasImage = True
                tag = image.tag(item, scale='tile')
            else:
                hasImage = False
                tag = None

            if has_variations:
                skuCode = None
                price = None
            else:
                varConf = None
                skuCode = item.Schema().getField('skuCode').get(item)
                price = item.Schema().getField('price').get(item)

            results.append(dict(item = item,
                                title = item.Title(),
                                description = item.Description(),
                                url = item.absolute_url(),
                                hasImage = hasImage,
                                imageTag = tag,
                                variants = None,
                                skuCode = skuCode,
                                price = price,
                                showPrice = item.getField('showPrice').get(item),
                                unit=item.getField('unit').get(item),
                                uid = item.UID(),
                                varConf = varConf,
                                hasVariations = has_variations,
                                selectable_dimensions = item.getSelectableDimensions()))
        return results

    def getVariationsConfig(self):
        """Returns the variation config for the item currently being viewed
        """
        context = aq_inner(self.context)
        variation_config = IVariationConfig(context)
        return variation_config

    def getVarDictsJSON(self):
        """Returns a JSON serialized dict with UID:varDict pairs, where UID
        is the ShopItem's UID and varDict is the item's variation dict.
        This is being used for the compact category view where inactive
        item variations must not be buyable.
        """
        varDicts = {}
        items = self.getItemDatas()
        for item in items:
            uid = item['uid']
            varConf = item['varConf']
            if varConf is not None:
                varDicts[uid] = dict(varConf.getVariationDict())
            else:
                varDicts[uid] = {}

            # Convert Decimals to Strings for serialization
            varDict = varDicts[uid]
            for vcode in varDict.keys():
                i = varDict[vcode]
                for k in i.keys():
                    val = i[k]
                    if isinstance(val, Decimal):
                        val = str(val)
                        i[k] = val

        return simplejson.dumps(varDicts)


class EditVariationsView(BrowserView):
    """View for editing ShopItem Variations
    """
    template = ViewPageTemplateFile('templates/edit_variations.pt')

    def __call__(self):
        """
        Self-submitting form that displays ShopItem Variations
        and updates them

        """
        form = self.request.form

        if form.get('remove_level'):
            variation_config = IVariationConfig(self.context)
            variation_config.remove_level()

        if form.get('reduce_level'):
            variation_config = IVariationConfig(self.context)
            variation_config.reduce_level()

        if form.get('add_level'):
            variation_config = IVariationConfig(self.context)
            variation_config.add_level()


        if form.get('addvalue'):
            fn = None
            idx_and_pos = form.get('addvalue')
            idx, pos = idx_and_pos.split('-')
            idx = int(idx)
            pos = int(pos) + 1

            if idx == 0:
                fn = 'variation1_values'
            elif idx == 1:
                fn = 'variation2_values'

            variation_config = IVariationConfig(self.context)
            values = list(self.context.getField(fn).get(self.context))
            var_dict = variation_config.getVariationDict()
            new_var_dict = {}
            pps = getMultiAdapter((self.context, self.request), name='plone_portal_state')
            language = pps.language()
            new_description = translate(_('label_new_description', default=u'New description'), domain='ftw.shop', context=self.context, target_language=language)
            DEFAULT_VARDATA = {'active':True, 'price': '0.00', 'skuCode': '99999', 'description': new_description}


            if len(variation_config.getVariationAttributes()) == 2:
                values1 = list(self.context.getField('variation1_values').get(self.context))
                values2 = list(self.context.getField('variation2_values').get(self.context))

                # Create a dict mapping old combination indexes to the new ones
                code_map = {}
                if idx == 0:
                    for i in range(len(values1)):
                        for j in range(len(values2)):
                            old_vcode = "var-%s-%s" % (i, j)
                            if i >= pos:
                                new_vcode = "var-%s-%s" % (i + 1, j)
                                code_map[old_vcode] = new_vcode
                            else:
                                code_map[old_vcode] = old_vcode
                elif idx == 1:
                    for i in range(len(values1)):
                        for j in range(len(values2)):
                            old_vcode = "var-%s-%s" % (i, j)
                            if j >= pos:
                                new_vcode = "var-%s-%s" % (i, j + 1)
                                code_map[old_vcode] = new_vcode
                            else:
                                code_map[old_vcode] = old_vcode

                # Based on the code map, reorder the var_dict
                for old_vcode in code_map.keys():
                    new_vcode = code_map[old_vcode]
                    new_var_dict[new_vcode] = var_dict[old_vcode]

                # Now add some default variation data for the value just added
                if idx == 0:
                    for j in range(len(values2)):
                        vcode = "var-%s-%s" % (pos, j)
                        new_var_dict[vcode] = DEFAULT_VARDATA
                elif idx == 1:
                    for i in range(len(values1)):
                        vcode = "var-%s-%s" % (i, pos)
                        new_var_dict[vcode] = DEFAULT_VARDATA


            elif len(variation_config.getVariationAttributes()) == 1:
                assert(idx == 0)
                values1 = list(self.context.getField('variation1_values').get(self.context))

                # Create a dict mapping old combination indexes to the new ones
                code_map = {}
                for i in range(len(values1)):
                    old_vcode = "var-%s" % (i)
                    if i >= pos:
                        new_vcode = "var-%s" % (i + 1)
                        code_map[old_vcode] = new_vcode
                    else:
                        code_map[old_vcode] = old_vcode

                # Based on the code map, reorder the var_dict
                for old_vcode in code_map.keys():
                    new_vcode = code_map[old_vcode]
                    new_var_dict[new_vcode] = var_dict[old_vcode]

                # Now add some default variation data for the value just added
                vcode = "var-%s" % pos
                new_var_dict[vcode] = DEFAULT_VARDATA


            # Finally purge and update the var_dict
            variation_config.purge_dict()
            variation_config.updateVariationConfig(new_var_dict)
            pps = getMultiAdapter((self.context, self.request), name='plone_portal_state')
            language = pps.language()
            new_value = translate(_('label_new_value', default=u'New value'), domain='ftw.shop', context=self.context, target_language=language)
            values.insert(pos, new_value)
            self.context.getField(fn).set(self.context, values)



        if form.get('delvalue'):
            fn = None
            idx_and_pos = form.get('delvalue')
            idx, pos = idx_and_pos.split('-')
            idx = int(idx)
            pos = int(pos)

            if idx == 0:
                fn = 'variation1_values'
            elif idx == 1:
                fn = 'variation2_values'
            values = list(self.context.getField(fn).get(self.context))


            variation_config = IVariationConfig(self.context)
            var_dict = variation_config.getVariationDict()
            new_var_dict = {}

            if len(variation_config.getVariationAttributes()) == 2:
                values1 = list(self.context.getField('variation1_values').get(self.context))
                values2 = list(self.context.getField('variation2_values').get(self.context))

                # Create a dict mapping old combination indexes to the new ones
                code_map = {}
                if idx == 0:
                    for i in range(len(values1)):
                        for j in range(len(values2)):
                            old_vcode = "var-%s-%s" % (i, j)
                            if i > pos:
                                new_vcode = "var-%s-%s" % (i - 1, j)
                                code_map[old_vcode] = new_vcode
                            elif i == pos:
                                code_map[old_vcode] = None
                            else:
                                code_map[old_vcode] = old_vcode
                elif idx == 1:
                    for i in range(len(values1)):
                        for j in range(len(values2)):
                            old_vcode = "var-%s-%s" % (i, j)
                            if j > pos:
                                new_vcode = "var-%s-%s" % (i, j - 1)
                                code_map[old_vcode] = new_vcode
                            elif j == pos:
                                code_map[old_vcode] = None
                            else:
                                code_map[old_vcode] = old_vcode

                # Based on the code map, reorder the var_dict
                for old_vcode in code_map.keys():
                    new_vcode = code_map[old_vcode]
                    new_var_dict[new_vcode] = var_dict[old_vcode]


            elif len(variation_config.getVariationAttributes()) == 1:
                assert(idx == 0)
                values1 = list(self.context.getField('variation1_values').get(self.context))

                # Create a dict mapping old combination indexes to the new ones
                code_map = {}
                for i in range(len(values1)):
                    old_vcode = "var-%s" % (i)
                    if i >= pos:
                        new_vcode = "var-%s" % (i - 1)
                        code_map[old_vcode] = new_vcode
                    else:
                        code_map[old_vcode] = old_vcode

                # Based on the code map, reorder the var_dict
                for old_vcode in code_map.keys():
                    new_vcode = code_map[old_vcode]
                    new_var_dict[new_vcode] = var_dict[old_vcode]


            # # Finally purge and update the var_dict
            variation_config.purge_dict()
            variation_config.updateVariationConfig(new_var_dict)

            values.pop(int(pos))
            self.context.getField(fn).set(self.context, values)




        if form.get('update_structure'):
            var_config = IVariationConfig(self.context)
            vattr0 = form.get('vattr-0', None)
            vattr1 = form.get('vattr-1', None)

            if len(var_config.getVariationAttributes()) >= 1:
                self.context.getField('variation1_attribute').set(self.context, vattr0)
                new_values = []
                for i in range(len(var_config.getVariation1Values())):
                    new_value = form.get('vvalue-%s-%s' % (0, i))
                    new_values.append(new_value)
                self.context.getField('variation1_values').set(self.context, new_values)

            if len(var_config.getVariationAttributes()) >= 2:
                self.context.getField('variation2_attribute').set(self.context, vattr1)
                new_values = []
                for j in range(len(var_config.getVariation2Values())):
                    new_value = form.get('vvalue-%s-%s' % (1, j))
                    new_values.append(new_value)
                self.context.getField('variation2_values').set(self.context, new_values)


        # Make sure we had a proper form submit, not just a GET request
        submitted = form.get('form.submitted', False)
        if submitted:
            variation_config = IVariationConfig(self.context)
            edited_var_data = self._parse_edit_variations_form()
            variation_config.updateVariationConfig(edited_var_data)

            IStatusMessage(self.request).addStatusMessage(
                _(u'msg_variations_saved',
                  default=u"Variations saved."), type="info")
            self.request.RESPONSE.redirect(self.context.absolute_url())

        return self.template()

    def _parse_edit_variations_form(self):
        """Parses the form the user submitted when editing variations,
        and returns a dictionary that contains the variation data.
        """
        form = self.request.form
        variation_config = IVariationConfig(self.context)
        variation_data = {}

        def _parse_data(variation_code):
            data = {}
            data['active'] = bool(form.get("%s-active" % variation_code))
            if data['active']:
                # TODO: Handle decimal more elegantly
                price = form.get("%s-price" % variation_code)
                try:
                    p = int(price)
                    # Create a tuple of ints from string
                    digits = tuple([int(i) for i in list(str(p))]) + (0, 0)
                    data['price'] = Decimal((0, digits, -2))
                except ValueError:
                    if not price == "":
                        data['price'] = Decimal(price)
                    else:
                        data['price'] = Decimal("0.00")

                data['skuCode'] = form.get("%s-skuCode" % variation_code)
                data['description'] = form.get("%s-description" % variation_code)

            # At this point the form has already been validated,
            # so uniqueness of sku codes is ensured
            data['hasUniqueSKU'] = True
            return data


        if len(variation_config.getVariationAttributes()) == 1:
            # One variation attribute
            for i, var1_value in enumerate(variation_config.getVariation1Values()):
                variation_code = 'var-%s' % i

                variation_data[variation_code] = _parse_data(variation_code)
        else:
            # Two variation attributes
            for i, var1_value in enumerate(variation_config.getVariation1Values()):
                for j, var2_value in enumerate(variation_config.getVariation2Values()):
                    variation_code = 'var-%s-%s' % (i, j)
                    variation_data[variation_code] = _parse_data(variation_code)

        return variation_data

    def getVariationsConfig(self):
        """Returns the variation config for the item being edited
        """
        context = aq_inner(self.context)
        variation_config = IVariationConfig(context)
        return variation_config


class MigrateVariationsView(BrowserView):
    """View to migrate variations of all shop items
    """

    def __call__(self):
        """Migrates all items
        """
        catalog = getToolByName(self.context, 'portal_catalog')
        normalize = getUtility(IIDNormalizer).normalize
        response = ""
        stats = {}

        items = catalog(portal_type="ShopItem")
        for item in items:
            obj = item.getObject()
            stats[obj.UID()] = {'status': 'UNKNOWN',
                                'result': 'UNKNOWN'}
            var_conf = IVariationConfig(obj)

            # Skip broken OrderedDict items
            var_dict = var_conf.getVariationDict()
            if str(type(var_dict)) == "<class 'zc.dict.dict.OrderedDict'>":
                status = "SKIPPED: Broken OrderedDict Item '%s' at '%s'" % (obj.Title(), obj.absolute_url())
                response += status + "\n"
                print status
                stats[obj.UID()] = {'status': 'SKIPPED',
                                    'result': 'SUCCESS'}
                continue

            varAttrs = var_conf.getVariationAttributes()
            num_variations = len(varAttrs)
            if num_variations == 0:
                # No migration needed
                stats[obj.UID()] = {'status': 'NO_MIGRATION_NEEDED',
                                    'result': 'SUCCESS'}
                continue

            # Migrate items with 2 variations
            if num_variations == 2:
                migrated = True

                # Create mapping from old to new keys
                mapping = {}
                for i, v1 in enumerate(var_conf.getVariation1Values()):
                    for j, v2 in enumerate(var_conf.getVariation2Values()):
                        vkey = "%s-%s" % (normalize(v1), normalize(v2))
                        vcode = "var-%s-%s" % (i, j)
                        mapping[vkey] = vcode

                # Check if item needs to be migrated
                for key in var_dict.keys():
                    if key in mapping.keys():
                        migrated = False

                if migrated:
                    # Already migrated
                    stats[obj.UID()] = {'status': 'ALREADY_MIGRATED',
                                        'result': 'SUCCESS'}
                else:
                    # Migrate the item
                    print "Migrating %s..." % obj.Title()
                    for vkey in mapping.keys():
                        vcode = mapping[vkey]
                        try:
                            # Store data with new vcode
                            var_dict[vcode] = var_dict[vkey]
                            del var_dict[vkey]
                            var_conf.updateVariationConfig(var_dict)
                            transaction.commit()
                            stats[obj.UID()] = {'status': 'MIGRATED',
                                                'result': 'SUCCESS'}
                        except KeyError:
                            status = "FAILED: Migration of item '%s' at '%s' failed!" % (obj.Title(), obj.absolute_url())
                            response += status + "\n"
                            print status
                            stats[obj.UID()] = {'status': 'FAILED',
                                                'result': 'FAILED'}
                            break
                    if stats[obj.UID()]['status'] == 'MIGRATED':
                        status = "MIGRATED: Item '%s' at '%s'" %  (obj.Title(), obj.absolute_url())
                        response += status + "\n"
                        print status


            # Migrate items with 1 variation
            if num_variations == 1:
                migrated = True

                # Create mapping from old to new keys
                mapping = {}
                for i, v1 in enumerate(var_conf.getVariation1Values()):
                    vkey = normalize(v1)
                    vcode = "var-%s" % i
                    mapping[vkey] = vcode

                # Check if item needs to be migrated
                for key in var_dict.keys():
                    if key in mapping.keys():
                        migrated = False


                if migrated:
                    # Already migrated
                    stats[obj.UID()] = {'status': 'ALREADY_MIGRATED',
                                        'result': 'SUCCESS'}
                else:
                    # Migrate this item
                    print "Migrating %s..." % obj.Title()
                    for vkey in mapping.keys():
                        vcode = mapping[vkey]
                        try:
                            # Store data with new vcode
                            var_dict[vcode] = var_dict[vkey]
                            del var_dict[vkey]
                            var_conf.updateVariationConfig(var_dict)
                            transaction.commit()
                            stats[obj.UID()] = {'status': 'MIGRATED',
                                                'result': 'SUCCESS'}
                        except KeyError:
                            status = "FAILED: Migration of item '%s' at '%s' failed!" % (obj.Title(), obj.absolute_url())
                            response += status + "\n"
                            print status
                            stats[obj.UID()] = {'status': 'FAILED',
                                                'result': 'FAILED'}
                            break
                    if stats[obj.UID()]['status'] == 'MIGRATED':
                        status = "MIGRATED: Item '%s' at '%s'" %  (obj.Title(), obj.absolute_url())
                        response += status + "\n"
                        print status



        total = len(items)
        migrated = len([stats[k] for k in stats if stats[k]['status'] == 'MIGRATED'])
        skipped = len([stats[k] for k in stats if stats[k]['status'] == 'SKIPPED'])
        failed = len([stats[k] for k in stats if stats[k]['status'] == 'FAILED'])
        no_migration_needed = len([stats[k] for k in stats if stats[k]['status'] == 'NO_MIGRATION_NEEDED'])
        already = len([stats[k] for k in stats if stats[k]['status'] == 'ALREADY_MIGRATED'])

        summary = "TOTAL: %s   MIGRATED: %s   SKIPPED: %s   "\
                  "FAILED: %s   NO MIGRATION NEEDED: %s   ALREADY_MIGRATED: %s" % (total,
                                                                          migrated,
                                                                          skipped,
                                                                          failed,
                                                                          no_migration_needed,
                                                                          already)
        response = "%s\n\n%s" % (summary, response)
        return response



from ftw.shop.tests.base import FtwShopTestCase
from pyquery import PyQuery as pq
from zope.component import getMultiAdapter
import unittest


class TestShopItemViews(FtwShopTestCase):

    def afterSetUp(self):
        super(TestShopItemViews, self).afterSetUp()

        self.movie_view = getMultiAdapter((self.movie,
                                 self.portal.REQUEST),
                                 name='view')
        self.book_view = getMultiAdapter((self.book,
                                 self.portal.REQUEST),
                                 name='view')
        self.tshirt_view = getMultiAdapter((self.tshirt,
                                 self.portal.REQUEST),
                                 name='view')

        self.book_edit_variations = getMultiAdapter((self.book,
                                 self.portal.REQUEST),
                                 name='edit_variations')

        self.tshirt_edit_variations = getMultiAdapter((self.tshirt,
                                 self.portal.REQUEST),
                                 name='edit_variations')


    def test_get_items(self):
        self.assertEquals(self.movie_view.getItems(),
                          [self.movie])

    def test_single_item(self):
        item_datas = self.movie_view.getItemDatas()
        item = item_datas[0]
        item_listing = self.movie_view.single_item(item)

        pq_doc = pq(item_listing)
        self.assertEquals(len(pq_doc("input[name=skuCode][value='12345']")), 1)

    def test_one_variation(self):
        item_datas = self.book_view.getItemDatas()
        item = item_datas[0]
        item_listing = self.book_view.one_variation(item)
        self.assertTrue('Paperback' in item_listing)
        self.assertTrue('Hardcover' in item_listing)

        pq_doc = pq(item_listing)
        self.assertEquals(pq_doc("#var-Hardcover td")[2].text, 'b11')
        self.assertEquals(pq_doc("#var-Paperback td")[2].text, 'b22')

    def test_two_variations(self):
        item_datas = self.tshirt_view.getItemDatas()
        item = item_datas[0]
        item_listing = self.tshirt_view.two_variations(item)
        pq_doc = pq(item_listing)
        self.assertEquals(len(pq_doc('select[name=var1choice] option')), 3)
        self.assertEquals(len(pq_doc('select[name=var2choice] option')), 3)

    def test_get_item_datas(self):
        movie_data = self.movie_view.getItemDatas()[0]
        book_data = self.book_view.getItemDatas()[0]
        tshirt_data = self.tshirt_view.getItemDatas()[0]

        self.assertEquals(movie_data['description'],
                          'A Shop Item with no variations')
        self.assertEquals(movie_data['hasVariations'], False)
        self.assertEquals(movie_data['imageTag'], None)
        self.assertEquals(movie_data['skuCode'], '12345')
        self.assertEquals(movie_data['price'], '7.15')
        self.assertEquals(movie_data['title'], 'A Movie')
        self.assertEquals(movie_data['varConf'], None)
        self.assertEquals(movie_data['variants'], None)

        self.assertEquals(book_data['description'],
                          'A Shop Item with one variation')
        self.assertEquals(book_data['hasVariations'], True)
        self.assertEquals(book_data['imageTag'], None)
        self.assertEquals(book_data['skuCode'], None)
        self.assertEquals(book_data['price'], None)
        self.assertEquals(book_data['title'], 'Professional Plone Development')
        self.assertTrue(book_data['varConf'])
        self.assertEquals(book_data['variants'], None)

        self.assertEquals(tshirt_data['description'],
                          'A Shop Item with two variations')
        self.assertEquals(tshirt_data['hasVariations'], True)
        self.assertEquals(tshirt_data['imageTag'], None)
        self.assertEquals(tshirt_data['skuCode'], None)
        self.assertEquals(tshirt_data['price'], None)
        self.assertEquals(tshirt_data['title'], 'A T-Shirt')
        self.assertTrue(tshirt_data['varConf'])
        self.assertEquals(tshirt_data['variants'], None)

    def test_get_variations_config(self):
        movie_vc = self.movie_view.getVariationsConfig()
        book_vc = self.book_view.getVariationsConfig()
        tshirt_vc = self.tshirt_view.getVariationsConfig()

        self.assertEquals(movie_vc.getVariationAttributes(), [])
        self.assertEquals(book_vc.getVariationAttributes(), ['Cover'])
        self.assertEquals(tshirt_vc.getVariationAttributes(), ['Color', 'Size'])


    def test_variations_order(self):
        var1_values = self.tshirt_vc.getVariation1Values()
        self.assertEquals(var1_values, ('Red', 'Green', 'Blue'))

        item_datas = self.tshirt_view.getItemDatas()
        item = item_datas[0]
        item_listing = self.tshirt_view.two_variations(item)

        pq_doc = pq(item_listing)
        self.assertEquals(
            ['Red', 'Green', 'Blue'],
            [op.text for op in pq_doc('select[name=var1choice] option')])

        # Now we swap 'Red' and 'Blue' and check if the listing reflects
        # the new order
        self.tshirt.Schema().getField('variation1_values').set(self.tshirt, ['Blue', 'Green', 'Red'])

        item_datas = self.tshirt_view.getItemDatas()
        item = item_datas[0]
        item_listing = self.tshirt_view.two_variations(item)

        pq_doc = pq(item_listing)
        self.assertEquals(
            ['Blue', 'Green', 'Red'],
            [op.text for op in pq_doc('select[name=var1choice] option')])

def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestShopItemViews))
    return suite



from zope.interface import implements
from zope.interface import Interface
from zope.component import adapts

from ftw.shop.interfaces import IStatusSet

from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary

from ftw.shop import shopMessageFactory as _

class DefaultStatusSet(object):
    implements(IStatusSet)
    adapts(Interface)
    # plone.registry.interfaces.IRecordsProxy
    # ftw.shop.interfaces.IShopConfiguration
    title = u"Default Status Set"

    vocabulary = SimpleVocabulary(
        [SimpleTerm(value=1,
                    token=1,
                    title=_(u'status_online_pending',
                            default=u'Pending (online payment)'),
                    ),
         SimpleTerm(value=2,
                    token=2,
                    title=_(u'status_confirmed',
                            default=u'Confirmed (online payment)'),
                    ),
         SimpleTerm(value=3,
                    token=3,
                    title=_(u'status_on_account',
                            default=u'On account'),
                    )
          ]
        )

    def __init__(self, context):
        self.context = context
        pass



from zope.component import adapts
from zope.interface import implements

from archetypes.schemaextender.field import ExtensionField
from archetypes.schemaextender.interfaces import ISchemaExtender
from Products.Archetypes import atapi

from Products.ATContentTypes.config import HAS_LINGUA_PLONE
if HAS_LINGUA_PLONE:
    from Products.LinguaPlone.public import BooleanField
    from Products.LinguaPlone.public import StringField
    from Products.LinguaPlone.public import LinesField
    from Products.LinguaPlone.public import FixedPointField
else:
    from Products.Archetypes.atapi import BooleanField
    from Products.Archetypes.atapi import StringField
    from Products.Archetypes.atapi import LinesField
    from Products.Archetypes.atapi import FixedPointField

from ftw.shop import shopMessageFactory as _
from ftw.shop.interfaces import IShopItem


class ExtBooleanField(ExtensionField, BooleanField):
    """A boolean field."""


class ExtStringField(ExtensionField, StringField):
    """A string field."""


class ExtFixedPointField(ExtensionField, FixedPointField):
    """A fixed point field."""


class ExtLinesField(ExtensionField, LinesField):
    """A lines field."""


class ShopItemExtender(object):
    """Extends the base type ShopItem with fields `price`
    and `skuCode`.
    """
    implements(ISchemaExtender)
    adapts(IShopItem)


    fields = [
        ExtFixedPointField('price',
            default = "0.00",
            required = 0,
            languageIndependent=True,
            widget = atapi.DecimalWidget(
                label = _(u"label_price", default=u"Price"),
                description = _(u"desc_price", default=u""),
                size=8,
            ),
        ),

        ExtBooleanField('showPrice',
            default = False,
            languageIndependent=True,
            widget = atapi.BooleanWidget(
                label = _(u"label_show_price", default=u"Show price"),
                description = _(u"desc_show_price", default=u""),
            ),
        ),

        ExtStringField('skuCode',
            required = 0,
            languageIndependent=True,
            widget = atapi.StringWidget(
                label = _(u"label_sku_code", default=u"SKU code"),
                description = _(u"desc_sku_code", default=u""),
            ),
        ),

        ExtStringField('variation1_attribute',
            required = 0,
            widget = atapi.StringWidget(
                label = _(u"label_variation1_attr",
                          default=u"Variation 1 Attribute"),
                description = _(u"desc_variation1_attr", default=u""),
                visible={'view': 'invisible', 'edit': 'invisible'},
            ),
        ),

        ExtLinesField('variation1_values',
            required = 0,
            widget = atapi.LinesWidget(
                label = _(u"label_variation1_values",
                          default=u"Variation 1 Values"),
                description = _(u"desc_variation1_values", default=u""),
                visible={'view': 'invisible', 'edit': 'invisible'},
            ),
        ),


        ExtStringField('variation2_attribute',
            required = 0,
            widget = atapi.StringWidget(
                label = _(u"label_variation2_attr",
                          default=u"Variation 2 Attribute"),
                description = _(u"desc_variation2_attr", default=u""),
                visible={'view': 'invisible', 'edit': 'invisible'},
            ),
        ),


        ExtLinesField('variation2_values',
            required = 0,
            widget = atapi.LinesWidget(
                label = _(u"label_variation2_values",
                          default=u"Variation 2 Values"),
                description = _(u"desc_variation2_values", default=u""),
                visible={'view': 'invisible', 'edit': 'invisible'},
            ),
        ),

        ExtFixedPointField('vat',
            default = "0.0",
            required = 0,
            languageIndependent=True,
            widget = atapi.SelectionWidget(
                label = _(u"label_vat_rate", default=u"VAT rate"),
                description = _(u"desc_vat_rate", default=u"Please select the value-added tax rate for this item."),
                size=8,
                format='select',
            ),
            vocabulary_factory = 'ftw.shop.VATRatesVocabulary',
        ),


    ]

    def __init__(self, context):
        self.context = context

    def getFields(self):
        return self.fields



from Products.ATContentTypes.content import base
from Products.ATContentTypes.content.schemata import ATContentTypeSchema
from Products.Archetypes import atapi
from ftw.shop import shopMessageFactory as _
from ftw.shop.config import PROJECTNAME
from ftw.shop.interfaces import ISupplier
from zope.interface import implements


from Products.ATContentTypes.config import HAS_LINGUA_PLONE
if HAS_LINGUA_PLONE:
    from Products.LinguaPlone.public import registerType
else:
    from Products.Archetypes.atapi import registerType


SupplierSchema = ATContentTypeSchema.copy() + atapi.Schema((

        atapi.TextField(
            name='email',
            required=True,
            label=_(u'label_supplier_email', default=u"E-Mail")),

        atapi.TextField(
            name='address',
            required=False,
            allowable_content_types=('text/plain', ),
            default_content_type='text/plain',
            default_input_type='text/plain',
            default_output_type='text/x-html-safe',

            widget=atapi.RichWidget(
                label=_(u'label_supplier_address', default=u"Address"),
                description=_(u'help_supplier_address',
                              default=u"Supplier's postal address")))))


SupplierSchema['title'].widget.label = _(u"label_name", default="Name")
SupplierSchema['title'].widget.description = _(u"help_name",
                                               default="The supplier's name")

SupplierSchema['description'].widget.visible = {'edit': 'invisible',
                                                'view': 'invisible'}


class Supplier(base.ATCTContent):
    """A supplier
    """
    implements(ISupplier)

    meta_type = "Supplier"
    schema = SupplierSchema

registerType(Supplier, PROJECTNAME)



import unittest
import zope.event
from Products.Archetypes.event import ObjectInitializedEvent
from ftw.shop.tests.base import FtwShopTestCase


class TestSetup(FtwShopTestCase):

    def afterSetUp(self):
        super(TestSetup, self).afterSetUp()

    def test_add_shop_type_permissisons(self):
        # The API of the permissionsOfRole() function sucks - it is bound too
        # closely up in the permission management screen's user interface
        roles = ['Manager', 'Contributor']
        for r in roles:
            selected_permissions = [p['name'] for p in
                                    self.portal.permissionsOfRole(r) if p['selected']]
            self.failUnless('ftw.shop: Add Shop Item' in selected_permissions)
            self.failUnless('ftw.shop: Add Shop Category' in selected_permissions)

    def test_shop_types_installed(self):
        self.failUnless('ShopCategory' in self.types.objectIds())
        self.failUnless('ShopItem' in self.types.objectIds())

    def test_shop_category_fti(self):
        document_fti = getattr(self.types, 'ShopCategory')
        self.failUnless(document_fti.global_allow)
        self.failUnless('ShopCategory' in document_fti.allowed_content_types)
        self.failUnless('ShopItem' in document_fti.allowed_content_types)

    def test_shop_item_creation(self):
        self.setRoles(('Manager',))
        self.portal.shop.products.invokeFactory('ShopItem', 'test-item')
        self.setRoles(('Member',))
        self.failUnless(self.portal.shop.products['test-item'].id == 'test-item')

    def test_shop_item_default_category(self):
        self.setRoles(('Manager',))
        self.portal.shop.products.invokeFactory('ShopItem', 'test-item')
        item = self.portal.shop.products['test-item']
        self.setRoles(('Member',))

        event = ObjectInitializedEvent(item, self.portal.REQUEST)
        zope.event.notify(event)

        self.failUnless(self.portal.shop.products in item.listCategories())



def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestSetup))
    return suite



from ftw.builder import Builder
from ftw.builder import create
from ftw.shop.interfaces import IShopRoot
from ftw.shop.testing import FTW_SHOP_FUNCTIONAL_TESTING
from ftw.shop.tests.pages import manage_categories
from ftw.testbrowser import browsing
from plone.app.testing import login
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from unittest2 import TestCase
import transaction


class TestBrowserManageCategories(TestCase):

    layer = FTW_SHOP_FUNCTIONAL_TESTING

    def setUp(self):
        portal = self.layer['portal']
        setRoles(portal, TEST_USER_ID, ['Manager'])
        login(portal, TEST_USER_NAME)

        shop_folder = create(Builder('folder')
                             .titled('shop')
                             .providing(IShopRoot))

        category_1 = create(Builder('shop category')
                            .within(shop_folder)
                            .titled('Category 1'))

        create(Builder('shop category')
               .within(category_1)
               .titled('Sub Category 1.1'))

        create(Builder('shop category')
               .within(category_1)
               .titled('Sub Category 1.2'))

        category_2 = create(Builder('shop category')
                            .within(shop_folder)
                            .titled('Category 2'))

        create(Builder('shop category')
               .within(category_2)
               .titled('Sub Category 2.1'))

        self.shop_item= create(Builder('shop item').within(category_1))

        transaction.commit()

    @browsing
    def test_manage_categories_shows_category_tree(self, browser):
        browser.login(SITE_OWNER_NAME).open()
        manage_categories.visit_edit_categories(self.shop_item)
        category_tree = manage_categories.review_category_tree()
        self.assertEquals(
        [
            [
                ['Category 1', ''],
                ['Sub Category 1.1', 'Category 1 > Sub Category 1.1'],
                ['Sub Category 1.2', 'Category 1 > Sub Category 1.2']
            ],
            [
                ['Category 2', ''],
                ['Sub Category 2.1', 'Category 2 > Sub Category 2.1']
            ]
        ],
        category_tree)



from Acquisition import aq_inner, aq_parent
from collections import OrderedDict
from decimal import Decimal
from decimal import getcontext
from decimal import InvalidOperation
from decimal import ROUND_UP
from ftw.shop import shopMessageFactory as _
from ftw.shop.config import CART_KEY
from ftw.shop.config import SESSION_ADDRESS_KEY, ONACCOUNT_KEY
from ftw.shop.exceptions import MissingCustomerInformation
from ftw.shop.exceptions import MissingOrderConfirmation
from ftw.shop.exceptions import MissingPaymentProcessor
from ftw.shop.exceptions import MissingShippingAddress
from ftw.shop.interfaces import IPaymentProcessor
from ftw.shop.interfaces import IShopConfiguration
from ftw.shop.interfaces import IShoppingCart
from ftw.shop.interfaces import IVariationConfig
from plone import api
from plone.app.layout.navigation.interfaces import INavigationRoot
from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from zope.component import adapts
from zope.component import getAdapters
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.interface import implements
from zope.interface import Interface
from zope.publisher.interfaces.browser import IBrowserRequest
import simplejson


class ShoppingCartAdapter(object):
    """Adapter that represents the shopping cart for the current user, defined
    by the adapted context and request.

    This adapter is responsible for adding, updating and removing items from the
    cart and storing that information in the user's session, as well as
    calculating the current cart total.
    """
    implements(IShoppingCart)
    adapts(Interface, IBrowserRequest)

    def __init__(self, context, request):
        self.context = context
        self.request = request

        # Setup Decimal environment
        getcontext().rounding = ROUND_UP

    @property
    def catalog(self):
        if not hasattr(self, '_catalog'):
            self._catalog = getToolByName(self.context, 'portal_catalog')
        return self._catalog

    def get_item_key(self, uid, variation_code, dimensions=None):
        """A unique id for each item / variation / dimension combination.
        """
        dimcode = '-'.join([str(dim) for dim in dimensions]) if dimensions else ''

        return '='.join([uid, variation_code, dimcode])

    def get_items(self):
        """Get all items currently contained in shopping cart
        """
        # Avoid creating a session if there doesn't exist one yet.
        bid_manager = getToolByName(self.context, 'browser_id_manager')
        browser_id = bid_manager.getBrowserId(create=0)
        if not browser_id:
            return {}
        session = self.request.SESSION
        items = session.get(CART_KEY, OrderedDict())
        return items

    def get_vat(self):
        """Return the cart's total VAT as a string
        """
        items = self.get_items()
        vat_total = Decimal('0.00')
        for item in items.values():
            vat_total += Decimal(item['vat_amount'])
        return str(vat_total)

    def get_total(self):
        """Return the cart's total as a string
        """
        items = self.get_items()
        total = Decimal('0.00')
        for item in items.values():
            total += Decimal(item['total'])
        return '{:.2f}'.format(total)

    def _get_supplier_info(self, context):
        supplier = self._get_supplier(context)
        if supplier:
            supplier_name = supplier.getField('title').get(supplier)
            supplier_email = supplier.getField('email').get(supplier)
            return supplier_name, supplier_email
        return ('', '')

    def _get_supplier(self, context):
        """If available, get supplier name and email address
        """
        supplier = None

        while not INavigationRoot.providedBy(context):
            field = context.getField('supplier')
            if field is not None:
                supplier = field.get(context)

                if supplier:
                    break

            context = aq_parent(aq_inner(context))

        return supplier

    def get_suppliers(self):
        suppliers = []

        for itemkey, itemvalue in self.get_items().items():
            if 'variation_code' in itemvalue:
                # Strip of variation suffix from itemkey to get the uid.
                variation_suffix = itemvalue['variation_code'].strip('var')
                uid = itemkey[:-len(variation_suffix)]
            else:
                uid = itemkey

            item = self.catalog(UID=uid)[0].getObject()
            supplier = self._get_supplier(item)
            suppliers.append(supplier)
        return suppliers

    def get_show_prices(self):
        for item in self.get_items().values():
            if item['show_price']:
                return True
        return False

    def calc_item_total(self, price, quantity, dimensions, price_modifier):
        # calculates the item total form price, dimensions and quantity
        # the dimensions_modifier is needed because the price is sometimes in a
        # different unit than the dimensions (amount in g, price in kg)
        return quantity * self.calc_price_per_item(price, dimensions, price_modifier)

    def calc_price_per_item(self, price, dimensions, price_modifier):
        size = reduce(lambda x, y: x * y, dimensions) if dimensions else 1
        return Decimal(size) * Decimal(price) / Decimal(price_modifier)

    def calc_vat(self, rate, total):
        """Calculate VAT and round to correct precision.
        """
        vat_amount = (Decimal(rate) / 100) * total
        vat_amount = vat_amount.quantize(Decimal('.01'))
        return vat_amount

    def has_single_supplier(self):
        """Returns true if all the suppliers are the same, or there are no
        suppliers whatsoever.
        """
        suppliers = self.get_suppliers()
        return all(x == suppliers[0] for x in suppliers)

    def has_specified_dimensions(self):
        """Checks if any item has dimensions.
        """
        for item in self.get_items().values():
            if len(item['dimensions']) > 0:
                return True
        return False

    def update_item(self, key, quantity, dimensions):
        """Update the quantity or dimensions of an item.
        """
        session = self.request.SESSION
        cart_items = session.get(CART_KEY, OrderedDict())

        if key not in cart_items:
            return

        item = cart_items[key]
        item['quantity'] = int(quantity)
        item['dimensions'] = dimensions
        price_per_item = self.calc_price_per_item(Decimal(item['price']),
                                                  dimensions,
                                                  item['price_modifier'])
        total = self.calc_item_total(Decimal(item['price']),
                                     item['quantity'],
                                     dimensions,
                                     item['price_modifier'])
        item['price_per_item'] = '{:.2f}'.format(price_per_item)
        item['total'] = '{:.2f}'.format(total)
        item['vat_amount'] = str(self.calc_vat(item['vat_rate'], total))
        cart_items[key] = item

        # update the key if the dimensions changed
        new_key = self.get_item_key(
            item['uid'],
            item['variation_code'] if 'variation_code' in item else '',
            dimensions)
        if key != new_key:
            cart_items[new_key] = cart_items[key]
            del cart_items[key]

        session[CART_KEY] = cart_items

    def _add_item(self, skuCode=None, quantity=1,
                   var1choice=None, var2choice=None, dimensions=None):
        """ Add item to cart
        """
        context = aq_inner(self.context)
        varConf = IVariationConfig(self.context)

        # get current items in cart
        session = self.request.SESSION
        cart_items = session.get(CART_KEY, OrderedDict())
        variation_code = varConf.variation_code(var1choice, var2choice)

        # We got no skuCode, so look it up by variation key
        if not skuCode:
            skuCode = varConf.sku_code(var1choice, var2choice)

        item_key = self.get_item_key(self.context.UID(), variation_code, dimensions)
        item = cart_items.get(item_key, None)

        item_title = context.Title()
        quantity = int(quantity)
        vat_rate = Decimal(context.getField('vat').get(context))

        supplier_name, supplier_email = self._get_supplier_info(context)

        if dimensions is None:
            dimensions = []

        price_modifier = self.context.getPriceModifier()

        has_variations = varConf.hasVariations()
        if has_variations:
            variation_dict = varConf.getVariationDict()
            if not variation_dict[variation_code]['active']:
                # Item is disabled
                return False

            variation_pretty_name = varConf.getPrettyName(variation_code)
            item_title = '%s - %s' % (context.Title(), variation_pretty_name)
            price = Decimal(variation_dict[variation_code]['price'])

            price_per_item = self.calc_price_per_item(price, dimensions, price_modifier)

            # add item to cart
            if item is None:
                total = self.calc_item_total(price, quantity, dimensions, price_modifier)

                item = {'uid': self.context.UID(),
                        'title': item_title,
                        'description': context.Description(),
                        'skucode': skuCode,
                        'quantity': quantity,
                        'price': str(price),
                        'show_price': context.showPrice,
                        'total': '{:.2f}'.format(total),
                        'price_per_item': '{:.2f}'.format(price_per_item),
                        'url': context.absolute_url(),
                        'supplier_name': supplier_name,
                        'supplier_email': supplier_email,
                        'variation_code': variation_code,
                        'vat_rate': vat_rate,
                        'vat_amount': str(self.calc_vat(vat_rate, total)),
                        'dimensions': dimensions,
                        'selectable_dimensions': context.getSelectableDimensions(),
                        'price_modifier': price_modifier
                }
            # item already in cart, update quantity
            else:
                item['quantity'] = item.get('quantity', 0) + quantity
                total = self.calc_item_total(price, item['quantity'], dimensions, price_modifier)
                item['dimensions'] = dimensions
                item['total'] = '{:.2f}'.format(total)
                item['price_per_item'] = '{:.2f}'.format(price_per_item)
                item['vat_amount'] = str(self.calc_vat(vat_rate, total))

        else:
            price = Decimal(context.Schema().getField('price').get(context))
            price_per_item = self.calc_price_per_item(price, dimensions, price_modifier)

            # add item to cart
            if item is None:
                total = self.calc_item_total(price, quantity, dimensions, price_modifier)
                item = {'uid': self.context.UID(),
                        'title': item_title,
                        'description': context.Description(),
                        'skucode': skuCode,
                        'quantity': quantity,
                        'price': str(price),
                        'show_price': context.showPrice,
                        'total': '{:.2f}'.format(total),
                        'price_per_item': '{:.2f}'.format(price_per_item),
                        'url': context.absolute_url(),
                        'supplier_name': supplier_name,
                        'supplier_email': supplier_email,
                        'vat_rate': vat_rate,
                        'vat_amount': str(self.calc_vat(vat_rate, total)),
                        'dimensions': dimensions,
                        'selectable_dimensions': context.getSelectableDimensions(),
                        'price_modifier': price_modifier
                }
            # item already in cart, update quantity
            else:
                item['quantity'] = item.get('quantity', 0) + quantity
                total = self.calc_item_total(price, item['quantity'], dimensions, price_modifier)
                item['dimensions'] = dimensions
                item['total'] = '{:.2f}'.format(total)
                item['price_per_item'] = '{:.2f}'.format(price_per_item)
                item['vat_amount'] = str(self.calc_vat(vat_rate, total))

        # store cart in session
        cart_items[item_key] = item
        session[CART_KEY] = cart_items
        return True

    def add_item(self, skuCode=None, quantity=1,
                 var1choice=None, var2choice=None, dimensions=None):
        """Add item to cart and redirect to referer.

        The item must be identified by either its skuCode if it is an item
        without variations, or by its variation key.
        """
        context = aq_inner(self.context)
        ptool = getToolByName(context, 'plone_utils')

        # Add item to cart
        success = self._add_item(skuCode, quantity, var1choice, var2choice, dimensions)
        if success:
            ptool.addPortalMessage(_(u'msg_item_added',
                                     default=u'Added item to cart.'), 'info')
        else:
            ptool.addPortalMessage(_(u'msg_item_disabled',
                                     default=u'Item is disabled and can\'t be added.'), 'error')

        # redirect to referer
        referer = self.request.get('HTTP_REFERER', context.absolute_url())
        if referer == 'localhost':
            referer = context.absolute_url()
        self.request.response.redirect(referer)

    def _remove_item(self, key):
        """Remove the item with the given key from the cart.
        """
        session = self.request.SESSION
        cart_items = session.get(CART_KEY, OrderedDict())

        if key in cart_items:
            del cart_items[key]

        session[CART_KEY] = cart_items

    def remove_item(self):
        """Remove an item from cart.
        """
        context = aq_inner(self.context)
        ptool = getToolByName(context, 'plone_utils')
        # remove_item doesn't expect skuCode but item keys - rename!
        skuCode = self.request.get('skuCode')
        if skuCode:
            self._remove_item(skuCode)
            ptool.addPortalMessage(_(u'msg_cart_updated',
                                     default=u"Cart updated."), 'info')

        referer = self.request.get('HTTP_REFERER', context.absolute_url())
        self.request.response.redirect(referer)
        return

    def purge_cart(self):
        """Remove all items from cart.
        """
        session = self.request.SESSION
        session[CART_KEY] = OrderedDict()

        context = aq_inner(self.context)
        ptool = getToolByName(context, 'plone_utils')
        ptool.addPortalMessage(_(u'msg_cart_emptied',
                                 default=u"Cart emptied."), 'info')
        referer = self.request.get('HTTP_REFERER', context.absolute_url())
        self.request.response.redirect(referer)
        return

    def update_cart(self):
        """Update cart contents.
        """
        context = aq_inner(self.context)
        ptool = getToolByName(context, 'plone_utils')

        # delete items with quantity 0
        del_items = []
        for item_key in self.get_items().keys():
            try:
                qty = int(float(self.request.get('quantity_%s' % item_key)))
                if qty <= 0:
                    del_items.append(item_key)
            except (ValueError, TypeError):
                ptool.addPortalMessage(
                    _(u'msg_cart_invalidvalue',
                      default=u"Invalid Values specified. Cart not updated."),
                    'error')
                referer = self.request.get('HTTP_REFERER',
                                           context.absolute_url())
                self.request.response.redirect(referer)
                return
        for item_key in del_items:
            self._remove_item(item_key)

        # now update quantities and dimensions
        for item_key, item in self.get_items().items():
            quantity = float(self.request.get('quantity_%s' % item_key))

            dimensions = self.request.get('dimension_%s' % item_key, [])
            if not isinstance(dimensions, list):
                dimensions = [dimensions]

            if not validate_dimensions(dimensions, item['selectable_dimensions']):
                ptool.addPortalMessage(
                    _(u'msg_cart_invalidvalue',
                      default=u"Invalid Values specified. Cart not updated."),
                    'error')
                referer = self.request.get('HTTP_REFERER',
                                           context.absolute_url())
                self.request.response.redirect(referer)
                return

            dimensions = map(Decimal, dimensions)

            # check that dimension changes do not collide
            item = self.get_items()[item_key]
            new_key = self.get_item_key(
                item['uid'],
                item['variation_code'] if 'variation_code' in item else '',
                dimensions)
            if new_key != item_key and new_key in self.get_items().keys():
                ptool.addPortalMessage(
                    _(u'msg_cart_invalidvalue',
                      default=u"Invalid Values specified. Cart not updated."),
                    'error')
                referer = self.request.get('HTTP_REFERER',
                                           context.absolute_url())
                self.request.response.redirect(referer)
                return

            self.update_item(item_key, quantity, dimensions)

        ptool.addPortalMessage(_(u'msg_cart_updated',
                                 default=u"Cart updated."), 'info')
        referer = self.request.get('HTTP_REFERER', context.absolute_url())
        self.request.response.redirect(referer)


class CartView(BrowserView):
    """Shopping Cart related helper view that takes care of tasks not done by
    the IShoppingCart adapter.

    For backwards compatibility, this view wraps many of the methods of said
    adapter for now, but will eventually be phased out.
    """

    portlet_template = ViewPageTemplateFile('../portlets/cart.pt')

    @property
    def cart(self):
        if not hasattr(self, '_cart'):
            self._cart = getMultiAdapter((self.context, self.request),
                                         IShoppingCart)
        return self._cart

    def addtocart_ajax(self, skuCode=None, quantity=1,
                       var1choice=None, var2choice=None, dimensions=None):
        """ Add item to cart, return portlet HTML and translated status
        message
        """
        translate = self.context.translate

        # unpack dimensions from request
        dimensions = dimensions.split('|') if dimensions else []

        if validate_dimensions(dimensions, self.context.getSelectableDimensions()):
            dimensions = map(Decimal, dimensions)

            success = self.cart._add_item(skuCode, quantity, var1choice, var2choice, dimensions)

            if success:
                status_msg_label = _(u'msg_label_info', default=u"Information")
                status_msg_text = _(u'msg_item_added', default=u"Added item to cart.")
            else:
                status_msg_label = _(u'msg_label_error', default=u"Error")
                status_msg_text = _(u'msg_item_disabled', default=u"Item is disabled and can't be added.")

        else:
            status_msg_label = _(u'msg_label_error', default=u"Error")
            status_msg_text = _(u'msg_invalid_dimensions', default=u"Invalid dimensions.")

        status_message = """\
<dt>%s</dt>
<dd>%s</dd>""" % (translate(status_msg_label),
               translate(status_msg_text))

        self.request.response.setHeader('Content-Type', 'application/json')
        return simplejson.dumps(dict(portlet_html=self.portlet_template(),
                    status_message=status_message))

    def addtocart(self, skuCode=None, quantity=1,
                  var1choice=None, var2choice=None, dimension=None):
        """Add item to cart and redirect to referer.

        The item must be identified by either its skuCode if it is an item
        without variations, or by its variation key.
        """
        # wrap single dimension in list so all dimensions are lists
        if not dimension:
            dimension = []
        if not isinstance(dimension, list):
            dimension = [dimension]

        if not validate_dimensions(dimension, self.context.getSelectableDimensions()):
            raise ValueError('Invalid dimensions.')

        dimension = map(Decimal, dimension)
        return self.cart.add_item(skuCode=skuCode, quantity=quantity,
                                  var1choice=var1choice, var2choice=var2choice,
                                  dimensions=dimension)

    def cart_items(self):
        """Get all items currently contained in shopping cart
        """
        return self.cart.get_items()

    def show_prices(self):
        return self.cart.get_show_prices()

    def cart_total(self):
        """Return the cart's total as a string
        """
        return self.cart.get_total()

    def cart_vat(self):
        """Return the cart's total VAT as a string
        """
        return self.cart.get_vat()

    def update_item(self, key, quantity, dimensions):
        """Update the quantity of an item.
        """
        return self.cart.update_item(key, quantity, dimensions)

    def cart_update(self):
        """Update cart contents.
        """
        return self.cart.update_cart()

    def cart_remove(self):
        """Remove an item from cart.
        """
        return self.cart.remove_item()

    def cart_delete(self):
        """Remove all items from cart.
        """
        return self.cart.purge_cart()

    def cart_has_dimensions(self):
        """Checks if any item has dimensions.
        """
        return self.cart.has_specified_dimensions()

    def checkout(self):
        """Process checkout
        """
        context = aq_inner(self.context)
        session = context.REQUEST.SESSION
        ptool = getToolByName(context, 'plone_utils')
        url = context.absolute_url()

        # check if we have something in the cart
        items = self.cart_items()
        if not items:
            ptool.addPortalMessage(_(u'msg_no_cart',
                            default=u"Can't proceed with empty cart."),
                'error')
            self.request.response.redirect(url)
        navroot = api.portal.get_navigation_root(context)
        omanager = getMultiAdapter((navroot, self.request),
                                   name=u'order_manager')


        # Check we got all the data we need from the wizard
        try:
            order_id = omanager.addOrder()
        except MissingCustomerInformation:
            self.request.response.redirect('%s/checkout-wizard' % url)
            return
        except MissingShippingAddress:
            self.request.response.redirect('%s/checkout-wizard' % url)
            return
        except MissingOrderConfirmation:
            self.request.response.redirect('%s/checkout-wizard' % url)
            return
        except MissingPaymentProcessor:
            self.request.response.redirect('%s/checkout-wizard' % url)
            return

        # Get the payment processor selected by the customer
        payment_processor_name = session.get(
            'payment_processor_choice').get('payment_processor')
        for name, adapter in getAdapters((context, context.REQUEST, context),
                                         IPaymentProcessor):
            if name == payment_processor_name:
                payment_processor = adapter

        if not payment_processor_name or not payment_processor.external:
            # No payment processor step at all OR payment by invoice

            customer_info = self.request.SESSION[SESSION_ADDRESS_KEY]
            self.request.SESSION.invalidate()

            # Get a new session object, since the old one still returns
            # stale data even though it has been invalidated
            session = context.session_data_manager.getSessionData()
            session[SESSION_ADDRESS_KEY] = customer_info

            # Set correct status for payment by invoice
            order = omanager.getOrder(order_id)
            order.status = ONACCOUNT_KEY

            omanager.sendOrderMails(order_id)

            self.request.response.redirect(
                        '%s/thankyou?order_id=%s' % (url, order.title))
            return
        else:
            #self.request.SESSION.invalidate()
            self.request.SESSION['order_id'] = order_id
            pp_launch_page = 'external-payment-processor'
            pp_launch_page = payment_processor.launch_page
            self.request.response.redirect('%s/%s' % (url, pp_launch_page))
            return

    def shop_config(self):
        """Return the shop configuration stored in the registry
        """
        registry = getUtility(IRegistry)
        shop_config = registry.forInterface(IShopConfiguration)
        return shop_config


def validate_dimensions(dimensions, selectable_dimensions):
    """ Checks if the request has the sane amount of dimensions. """
    if dimensions is None or selectable_dimensions is None:
        return False

    if len(dimensions) != len(selectable_dimensions):
        return False

    try:
        dimensions = map(Decimal, dimensions)
    except InvalidOperation:
        return False

    for dimension in dimensions:
        if dimension <= 0:
            return False

    return True



import unittest

from ftw.shop.content.categorizeable import Categorizeable
from ftw.shop.tests.base import FtwShopTestCase


class TestCategorizeable(FtwShopTestCase):
    
    def afterSetUp(self):
        super(TestCategorizeable, self).afterSetUp()

    def test_get_set_rank_for_category(self):
        mock_item = self.movie
        rank = mock_item.getRankForCategory(self.subcategory)
        self.assertEquals(rank, Categorizeable.defaultRank)
        
        mock_item.addToCategory(self.subcategory)
        mock_item.setRankForCategory(self.subcategory, 10)
        rank = mock_item.getRankForCategory(self.subcategory)
        self.assertEquals(rank, 10)
    
    def test_add_remove_category(self):
        mock_item = self.movie
        self.assertEquals(mock_item.listCategories(), [self.portal.shop.products])
        
        mock_item.addToCategory(self.subcategory)
        self.assertTrue(self.subcategory in mock_item.listCategories())
        
        mock_item.removeFromCategory(self.subcategory)
        self.assertTrue(self.subcategory not in mock_item.listCategories())



def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestCategorizeable))
    return suite



"""Test setup for integration and functional tests.

When we import PloneTestCase and then call setupPloneSite(), all of
Plone's products are loaded, and a Plone site will be created. This
happens at module level, which makes it faster to run each test, but
slows down test runner startup.
"""
from decimal import Decimal

from ftw.shop.interfaces import IMailHostAdapter
from ftw.shop.interfaces import IVariationConfig
from ftw.shop.mailer import MailHostAdapter

from Products.Archetypes.event import ObjectInitializedEvent
from Products.CMFCore.utils import getToolByName

from Products.Five import fiveconfigure
from Products.Five import zcml
from Products.PloneTestCase import PloneTestCase as ptc
from Products.PloneTestCase.layer import onsetup
from Testing import ZopeTestCase as ztc
from zope.component import adapts
from zope.component import getMultiAdapter
from zope.component import provideAdapter
from zope.interface import implements
from zope.interface import Interface
import zope.event


# Mock data to be used in various tests in ftw.shop but also
# other ftwshop.* products

MOCK_CUSTOMER = {'title': u'Mr.',
                 'firstname': u'Hugo',
                 'lastname': u'Boss',
                 'company': u'ACME Corp.',
                 'street1': u'Teststreet 23',
                 'street2': u'',
                 'zipcode': u'56789',
                 'city': u'Exampletown',
                 'email': u'hugo@example.org',
                 'phone': u'099 999 99 99',
                 'country': u'Switzerland'}

MOCK_SHIPPING = {'title': u'Mr.',
                'firstname': u'Hugo',
                'lastname': u'Boss',
                'company': u'ACME Corp.',
                'street1': u'Shippingstreet 42',
                'street2': u'',
                'zipcode': u'4242',
                'city': u'Exampletown'}


MOCK_CART = {'some-uid': {'description': 'A Shop Item with no variations',
                          'price': '4.15',
                          'quantity': 2,
                          'show_price': False,
                          'skucode': '12345',
                          'supplier_email': 'supplier@example.org',
                          'supplier_name': 'Supplier Name',
                          'title': 'Item Title',
                          'total': '16.60',
                          'url': 'http://nohost/plone/shop/products/item',
                          'vat_amount': '0.00',
                          'vat_rate': Decimal('0.00'),
                          'dimensions': [2],
                          'selectable_dimensions': [u'Weight (g)'],
                          'price_per_item': '8.30'}}

MOCK_CART_TWO_SUPPLIERS = {
    'other-uid': {'description': 'A Shop Item with no variations',
                  'price': '4.15',
                  'quantity': 2,
                  'show_price': False,
                  'skucode': '12345',
                  'supplier_email': 'supplier@example.org, other@example.org',
                  'supplier_name': 'Supplier Name',
                  'title': 'Item Title',
                  'total': '8.30',
                  'url': 'http://nohost/plone/shop/products/item',
                  'vat_amount': '0.00',
                  'vat_rate': Decimal('0.00'),
                  'dimensions': [],
                  'selectable_dimensions': [],
                  'price_per_item': '8.30'}}


class FakeMailHostAdapter(MailHostAdapter):
    """Fake MailHost Adapter used in tests.
    """
    implements(IMailHostAdapter)
    adapts(Interface)

    def __init__(self, context):
      from Products.CMFPlone.tests.utils import MockMailHost
      self.context = context
      mockmailhost = MockMailHost('MailHost')
      self.context.MailHost = mockmailhost


# When ZopeTestCase configures Zope, it will *not* auto-load products
# in Products/. Instead, we have to use a statement such as:
#   ztc.installProduct('SimpleAttachment')
# This does *not* apply to products in eggs and Python packages (i.e.
# not in the Products.*) namespace. For that, see below.
# All of Plone's products are already set up by PloneTestCase.

@onsetup
def setup_product():
    """Set up the package and its dependencies.

    The @onsetup decorator causes the execution of this body to be
    deferred until the setup of the Plone site testing layer. We could
    have created our own layer, but this is the easiest way for Plone
    integration tests.
    """

    # Load the ZCML configuration for the example.tests package.
    # This can of course use <include /> to include other packages.

    fiveconfigure.debug_mode = True
    import ftw.shop

    zcml.load_config('configure.zcml', ftw.shop)

    fiveconfigure.debug_mode = False

    # We need to tell the testing framework that these products
    # should be available. This can't happen until after we have loaded
    # the ZCML. Thus, we do it here. Note the use of installPackage()
    # instead of installProduct().
    # This is *only* necessary for packages outside the Products.*
    # namespace which are also declared as Zope 2 products, using
    # <five:registerPackage /> in ZCML.

    # We may also need to load dependencies, e.g.:
    #   ztc.installPackage('borg.localrole')

    ztc.installPackage('ftw.shop')


# The order here is important: We first call the (deferred) function
# which installs the products we need for this product. Then, we let
# PloneTestCase set up this product on installation.

setup_product()
ptc.setupPloneSite(products=['ftw.shop'])

class FtwShopTestCase(ptc.PloneTestCase):
    """We use this base class for all the tests in this package. If
    necessary, we can put common utility or setup code in here. This
    applies to unit test cases.
    """

    def afterSetUp(self):
        provideAdapter(FakeMailHostAdapter,
                 (Interface, ),
                 IMailHostAdapter)

        # Set up sessioning objects
        ztc.utils.setupCoreSessions(self.app)

        # Create an initial browser_id by requesting it
        bid_manager = getToolByName(self.app, 'browser_id_manager')
        bid_manager.getBrowserId()

        self.workflow = getToolByName(self.portal, 'portal_workflow')
        self.acl_users = getToolByName(self.portal, 'acl_users')
        self.types = getToolByName(self.portal, 'portal_types')

        self.setRoles(('Manager',))

        # Use helper method to set up shop root at portal.shop
        init_shop = getMultiAdapter((self.portal, self.portal.REQUEST),
                                    Interface, 'initialize-shop-structure')
        init_shop()

        # Create a shop category
        self.portal.shop.invokeFactory("ShopCategory", "products")
        self.portal.shop.products.reindexObject()

        # Create a Shop Item with no variations
        self.portal.shop.products.invokeFactory('ShopItem', 'movie')
        self.movie = self.portal.shop.products['movie']
        self.movie.getField('skuCode').set(self.movie, "12345")
        self.movie.getField('price').set(self.movie, "7.15")
        self.movie.getField('showPrice').set(self.movie, True)
        self.movie.getField('title').set(self.movie, "A Movie")
        self.movie.getField('description').set(self.movie, "A Shop Item with no variations")
        self.movie.getField('selectable_dimensions').set(self.movie, 'length_width_mm_mm2')

        # Fire ObjectInitializedEvent to add item to containing category
        event = ObjectInitializedEvent(self.movie, self.portal.REQUEST)
        zope.event.notify(event)

        self.movie.reindexObject()

        self.movie_vc = IVariationConfig(self.movie)

        # Create a Shop Item with one variation
        self.portal.shop.products.invokeFactory('ShopItem', 'book')
        self.book = self.portal.shop.products['book']
        self.book.getField('title').set(self.book, 'Professional Plone Development')
        self.book.getField('description').set(self.book, 'A Shop Item with one variation')
        self.book.getField('variation1_attribute').set(self.book, 'Cover')
        self.book.getField('variation1_values').set(self.book, ['Hardcover', 'Paperback'])

        # Fire ObjectInitializedEvent to add item to containing category
        event = ObjectInitializedEvent(self.book, self.portal.REQUEST)
        zope.event.notify(event)

        self.book.reindexObject()

        self.book_vc = IVariationConfig(self.book)
        book_var_dict = {
        'var-Hardcover': {'active': True,
                          'price': Decimal('1.00'),
                          'skuCode': 'b11',
                          'description': 'A hard and durable cover',
                          'hasUniqueSKU': True},
        'var-Paperback': {'active': True,
                          'price': Decimal('2.00'),
                          'skuCode': 'b22',
                          'description': 'A less durable but cheaper cover',
                          'hasUniqueSKU': True},
        }
        self.book_vc.updateVariationConfig(book_var_dict)

        # Create a Shop Item with two variations
        self.portal.shop.products.invokeFactory('ShopItem', 't-shirt')
        self.tshirt = self.portal.shop.products['t-shirt']
        self.tshirt.getField('title').set(self.tshirt, 'A T-Shirt')
        self.tshirt.getField('description').set(self.tshirt, 'A Shop Item with two variations')
        self.tshirt.getField('variation1_attribute').set(self.tshirt, 'Color')
        self.tshirt.getField('variation1_values').set(self.tshirt, ['Red', 'Green', 'Blue'])
        self.tshirt.getField('variation2_attribute').set(self.tshirt, 'Size')
        self.tshirt.getField('variation2_values').set(self.tshirt, ['S', 'M', 'L'])

        # Fire ObjectInitializedEvent to add item to containing category
        event = ObjectInitializedEvent(self.tshirt, self.portal.REQUEST)
        zope.event.notify(event)

        self.tshirt.reindexObject()

        self.tshirt_vc = IVariationConfig(self.tshirt)
        tshirt_var_dict = {
        'var-Red-S': {'active': True, 'price': Decimal('1.00'), 'skuCode': '11', 'description': '', 'hasUniqueSKU': True},
        'var-Red-M': {'active': True, 'price': Decimal('2.00'), 'skuCode': '22', 'description': '', 'hasUniqueSKU': True},
        'var-Red-L': {'active': True, 'price': Decimal('3.00'), 'skuCode': '33', 'description': '', 'hasUniqueSKU': True},
        'var-Green-S': {'active': True, 'price': Decimal('4.00'), 'skuCode': '44', 'description': '', 'hasUniqueSKU': True},
        'var-Green-M': {'active': True, 'price': Decimal('5.00'), 'skuCode': '55', 'description': '', 'hasUniqueSKU': True},
        'var-Green-L': {'active': True, 'price': Decimal('6.00'), 'skuCode': '66', 'description': '', 'hasUniqueSKU': True},
        'var-Blue-S': {'active': True, 'price': Decimal('7.00'), 'skuCode': '77', 'description': '', 'hasUniqueSKU': True},
        'var-Blue-M': {'active': True, 'price': Decimal('8.00'), 'skuCode': '88', 'description': '', 'hasUniqueSKU': True},
        'var-Blue-L': {'active': True, 'price': Decimal('9.00'), 'skuCode': '99', 'description': '', 'hasUniqueSKU': True},
        }
        self.tshirt_vc.updateVariationConfig(tshirt_var_dict)

        # Create a subcategory below the shop root
        self.portal.shop.products.invokeFactory("ShopCategory", "subcategory")
        self.subcategory = self.portal.shop.products.subcategory

        # Fire ObjectInitializedEvent to add category to containing category
        event = ObjectInitializedEvent(self.subcategory, self.portal.REQUEST)
        zope.event.notify(event)

        self.setRoles(('Member',))


    class Session(dict):
        def set(self, key, value):
            self[key] = value

        def invalidate(self):
            self.clear()


    def _setup(self):
        ptc.PloneTestCase._setup(self)
        self.app.REQUEST['SESSION'] = self.Session()


class FunctionalTestCase(ptc.FunctionalTestCase):
    """We use this class for functional integration tests that use
    doctest syntax. Again, we can put basic common utility or setup
    code in here.
    """

    def afterSetUp(self):
        roles = ('Member', 'Contributor')
        self.portal.portal_membership.addMember('contributor',
                                                'secret',
                                                roles, [])



from decimal import Decimal
from decimal import InvalidOperation
import codecs
import cStringIO
import csv
import re


def to_decimal(number):
    """Since SQLite doesn't support Decimal fields, trim the float it
    returns to two decimal places and convert it to Decimal. If that
    fails, return the number as-is."""
    try:
        if float(number) == 0.0:
            return Decimal('0.00')
    except ValueError:
        pass

    try:
        if str(number).find('.') == -1:
            return Decimal("%s.00" % number)
        return Decimal(str(number)[:str(number).find('.') + 3])
    except InvalidOperation:
        return number


def is_email_valid(address):
    expr = re.compile(r'^[^@]*[^@\.]@[^@\.][^@]*\.[^@]+$', re.IGNORECASE)
    return bool(expr.match(address))


class UnicodeCSVWriter:
    """A CSV writer which will write rows to CSV file "f",
    which is encoded in the given encoding.
    """

    def __init__(self, f, dialect=csv.excel, encoding="utf-8", **kwds):
        # Redirect output to a queue
        self.queue = cStringIO.StringIO()
        self.writer = csv.writer(self.queue, dialect=dialect, **kwds)
        self.stream = f
        self.encoder = codecs.getincrementalencoder(encoding)()

    def writerow(self, row):
        for i, value in enumerate(row):
            if value is None:
                row[i] = ''
            else:
                try:
                    row[i] = str(value)
                except UnicodeEncodeError:
                    row[i] = value

        encoded_row = []
        for s in row:
            try:
                encoded_row.append(s.encode('utf-8'))
            except UnicodeDecodeError:
                # Already encoded
                encoded_row.append(s)

        self.writer.writerow(encoded_row)
        # Fetch UTF-8 output from the queue ...
        data = self.queue.getvalue()
        data = data.decode("utf-8")
        # ... and reencode it into the target encoding
        data = self.encoder.encode(data)
        # write to the target stream
        self.stream.write(data)
        # empty queue
        self.queue.truncate(0)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)



from ftw.builder import Builder
from ftw.builder import create
from ftw.shop.tests.pages import cartportlet
from ftw.shop.tests.pages import shopitem
from ftw.testbrowser import browser


CONTACT_INFORMATION = '1. Contact Information'
SHIPPING_ADDRESS = '2. Shipping Address'
PAYMENT_PROCESSOR = '3. Payment Processor'
ORDER_REVIEW = '4. Order Review'


STEPS = [CONTACT_INFORMATION,
         SHIPPING_ADDRESS,
         PAYMENT_PROCESSOR,
         ORDER_REVIEW]


def add_item_to_cart(item, amount=1):
    browser.visit(item)
    shopitem.add_to_cart(amount=amount)


def visit_checkout_with_one_item_in_cart(item=None):
    if item is None:
        item = create(Builder('shop item'))
    add_item_to_cart(item)
    cartportlet.order_link().click()
    import checkout
    return checkout


def current_step():
    return browser.css('ul.wizard-steps > li.selected').first.text


def assert_step(expected):
    current = current_step()
    assert current == expected, \
        'Expected to be on step "%s", but current step is "%s"' % (
        expected, current)


def next():
    browser.find('Next').click()
    import checkout
    return checkout


def back():
    browser.find('Back').click()
    import checkout
    return checkout


def finish():
    browser.find('Finish').click()
    import checkout
    return checkout


def fill_contact_info():
    assert_step(CONTACT_INFORMATION)
    browser.fill({u'Title': 'Sir',
                  u'First Name': 'Hugo',
                  u'Last Name': 'Boss',
                  u'Email': 'hugo@boss.com',
                  u'Street/No.': 'Example Street 15',
                  u'Phone number': '001 0101 0101 01',
                  u'Zip Code': '3000',
                  u'City': 'Bern',
                  u'Country': 'Switzerland'})
    import checkout
    return checkout


def submit_valid_contact_info():
    return fill_contact_info().next()


def submit_valid_shipping_address():
    # defaults are already valid
    assert_step(SHIPPING_ADDRESS)
    return next()


def select_valid_payment_processor():
    assert_step(PAYMENT_PROCESSOR)
    # XXX: "Gegen Rechnung" should be translated to english
    browser.fill({'Gegen Rechnung': 'ftw.shop.InvoicePaymentProcessor'})
    import checkout
    return checkout


def submit_valid_payment_processor():
    return select_valid_payment_processor().next()


def goto(target):
    if browser.document is None:
        visit_checkout_with_one_item_in_cart()

    current = current_step()
    if current == target:
        import checkout
        return checkout

    current_index = STEPS.index(current)
    target_index = STEPS.index(target)
    assert current_index < target_index, \
        'Can not goto("%s"), currently on "%s" and can not backwards.' % (
        target, current)

    if current == CONTACT_INFORMATION:
        submit_valid_contact_info()
        return goto(target)

    elif current == SHIPPING_ADDRESS:
        submit_valid_shipping_address()
        return goto(target)

    elif current == PAYMENT_PROCESSOR:
        submit_valid_payment_processor()
        return goto(target)

    raise Exception('Failed to goto("%s"): unkown error' % target)


def review_contact_information():
    assert_step(ORDER_REVIEW)
    return browser.css('.contact-information li').text


def review_shipping_address():
    assert_step(ORDER_REVIEW)
    return browser.css('.shipping-address li').text



from DateTime import DateTime
from decimal import Decimal

from AccessControl import ClassSecurityInfo
from persistent import Persistent
from Products.CMFCore.permissions import ModifyPortalContent
from Products.CMFCore.utils import getToolByName
from zope.interface import implements

try:
    # Plone >= 4.0
    from zope.component.hooks import getSite

except ImportError:
    # Plone < 4.0
    from zope.app.component.hooks import getSite



from BTrees.IOBTree import IOBTree
try:
    from BTrees.LOBTree import LOBTree
    BTreeImplementation = LOBTree
except ImportError:
    BTreeImplementation = IOBTree
from BTrees.Length import Length

from ftw.shop.interfaces import IOrderStorage
from ftw.shop.utils import to_decimal


security = ClassSecurityInfo()


class Order(Persistent):

    def __init__(self):
        self.order_id = None
        self.status = None
        self.date = None
        self.total = None
        self.vat_amount = None
        self.title = None

        self.customer_title = None
        self.customer_firstname = None
        self.customer_lastname = None
        self.customer_email = None
        self.customer_company = None
        self.customer_street1 = None
        self.customer_street2 = None
        self.customer_phone = None
        self.customer_zipcode = None
        self.customer_city = None
        self.customer_country = None
        self.customer_shipping_address = None
        self.customer_comments = None

        self.shipping_title = None
        self.shipping_firstname = None
        self.shipping_lastname = None
        self.shipping_company = None
        self.shipping_street1 = None
        self.shipping_street2 = None
        self.shipping_zipcode = None
        self.shipping_city = None

        self.cartitems = None

    def getTotal(self):
        """Since SQLite doesn't support Decimal fields, trim the float it
        returns to two decimal places and convert it to Decimal. If that
        fails, return the total as-is."""
        return to_decimal(self.total)

    def getLocalizedDate(self):
        """Returns the localized order date and time.
        """
        site = getSite()
        util = getToolByName(site, 'translation_service')
        return util.ulocalized_time(self.date, long_format=True,
                                    context=site, domain='plonelocales')


class CartItems(Persistent):
    """Cart items model.
    """

    def __init__(self):
        self.id = None
        self.order_id = None
        self.sku_code = None
        self.quantity = None
        self.title = None
        self.price = None
        self.show_price = False
        self.total = None
        self.supplier_name = None
        self.supplier_email = None
        self.vat_rate = None
        self.vat_amount = None
        self.dimensions = None
        self.selectable_dimensions = None
        self.price_per_item = None

        self.order = None

    def getPrice(self):
        """Return the price converted to a Decimal
        """
        return to_decimal(self.price)

    def getTotal(self):
        """Return the total converted to a Decimal
        """
        return to_decimal(self.total)


class BTreeOrderStorage(Persistent):
    implements(IOrderStorage)
    title = u"BTree Storage"

    def __init__(self):
        self._orderStorage = BTreeImplementation()
        self._orderCount = 0
        self._length = Length()

    def getOrder(self, id):
        return self._orderStorage[id]

    def getAllOrders(self):
        return list(self._orderStorage.values())

    def _addDataRow(self, value):
        """Adds a row of data to the internal storage
        """

        # Get id for order to be added and increment internal counter
        id = self._orderCount + 1
        self._orderCount += 1

        self._orderStorage[id] = value
        self._length.change(1)
        return id


    security.declareProtected(ModifyPortalContent, 'createOrder')

    def createOrder(self, status=None, date=None, customer_data=None,
                    shipping_data=None, cart_data=None, total=None):
        """ a wrapper for the _addDataRow method """

        new_order = Order()
        new_order.status = status
        new_order.date = date
        new_order.total = Decimal(total)

        order_id = self._addDataRow(new_order)
        new_order.order_id = order_id

        # calc order number
        now = DateTime()
        order_prefix = '%03d%s' % (now.dayOfYear() + 500, now.yy())
        order_number = '%s%04d' % (order_prefix, order_id)
        new_order.title = order_number

        for key in customer_data.keys():
            setattr(new_order, "customer_%s" % key, customer_data[key])

        for key in shipping_data.keys():
            setattr(new_order, "shipping_%s" % key, shipping_data[key])

        # store cart in order
        all_cart_items = []
        vat_amount_total = Decimal('0.0')
        for key in cart_data.keys():
            cart_items = CartItems()
            cart_items.sku_code = cart_data[key]['skucode']
            cart_items.quantity = cart_data[key]['quantity']
            cart_items.title = cart_data[key]['title']
            cart_items.price = Decimal(cart_data[key]['price'])
            cart_items.show_price = cart_data[key]['show_price']
            cart_items.total = Decimal(cart_data[key]['total'])
            cart_items.supplier_name = cart_data[key]['supplier_name']
            cart_items.supplier_email = cart_data[key]['supplier_email']
            cart_items.vat_rate = cart_data[key]['vat_rate']
            cart_items.vat_amount = Decimal(cart_data[key]['vat_amount'])
            cart_items.dimensions = cart_data[key]['dimensions']
            cart_items.selectable_dimensions = cart_data[key]['selectable_dimensions']
            cart_items.price_per_item = cart_data[key]['price_per_item']
            cart_items.order_id = order_id
            cart_items.order = new_order

            all_cart_items.append(cart_items)
            vat_amount_total += cart_items.vat_amount

        new_order.cartitems = all_cart_items
        new_order.vat_amount = vat_amount_total
        return order_id

    def cancelOrder(self, order_id):
        del self._orderStorage[order_id]
        self._length.change(-1)

    def flush(self):
        pass

    def getFieldNames(self):
        o = Order()
        field_names = [f for f in dir(o) if not (f.startswith('_') or \
                                                 f.startswith('get') or \
                                                 f == 'cartitems')]
        return field_names



from ftw.shop.interfaces import IShopRoot
from ftw.upgrade import UpgradeStep
from zope.interface import noLongerProvides


class RemoveShopRootInterface(UpgradeStep):
    """Remove IShopRoot from all objects providing it"""

    def __call__(self):
        catalog = self.getToolByName('portal_catalog')
        query = {'object_provides': IShopRoot.__identifier__}
        msg = 'Remove IShopRoot from objects providing it'
        for obj in self.objects(query, msg):
            noLongerProvides(obj, IShopRoot)
            catalog.reindexObject(obj, idxs=['object_provides'])

        if IShopRoot.providedBy(self.portal):
            noLongerProvides(self.portal, IShopRoot)



from ftw.upgrade import UpgradeStep


class RemoveExtendedView(UpgradeStep):
    """Remove extended view.
    """

    def __call__(self):
        self.install_upgrade_profile()
        for obj in self.objects({
                'portal_type': ['Folder', 'ShopItem', 'ShopCategory']},
                'Remove old view from shop containers.'):

            if obj.getLayout() == 'compact_view':
                del obj.layout



import logging
import socket

from zope.component import adapts
from zope.interface import implements
from zope.interface import Interface
from Products.MailHost.MailHost import MailHostError
from email import message_from_string
from email.Header import Header

try:
    from email.encoders import encode_quopri, encode_base64
except ImportError:
    # Python 2.4, and therefore Plone < 4
    pass

from ftw.shop.interfaces import IMailHostAdapter

logger = logging.getLogger('ftw.shop')


class MailHostAdapter(object):
    """Abstracts a MailHost and handles slightly different implementations
    in Plone 4 and Plone <= 3.
    """
    implements(IMailHostAdapter)
    adapts(Interface)

    def __init__(self, context):
        self.context = context

    def send(self, msg_body, mto, mfrom=None, mbcc=None, reply_to=None, subject=None,
             encode=None, immediate=False, charset=None, msg_type=None):
        """Sends a message defined at least by its message body and the
        destination address.
        """
        mhost = self.context.MailHost

        try:
            # Plone 4
            msg = message_from_string(msg_body.encode(charset, 'ignore'))
            if encode is None or encode in ["quoted-printable", "qp"]:
                encode_quopri(msg)
            else:
                encode_base64(msg)

            msg['BCC']= Header(mbcc)
            if reply_to:
                msg['Reply-To'] = Header(reply_to)

            mhost.send(msg,
                         mto=mto,
                         mfrom=mfrom,
                         subject=subject,
                         encode=encode,
                         immediate=immediate,
                         msg_type=msg_type,
                         charset=charset)
        except (TypeError, NameError):
            # Plone 3 or earlier
            subtype = msg_type.split('/')[1]

            mhost.secureSend(msg_body,
                             mto=mto,
                             mfrom=mfrom,
                             subject=subject,
                             mbcc=mbcc,
                             subtype=subtype,
                             charset=charset)

        except (MailHostError, socket.error), e:
            logger.error("sending mail with subject %s to %s failed: %s."
                         % (subject, mto, str(e)))
        return



from ftw.shop import shopMessageFactory as _
from ftw.shop.config import DEFAULT_VAT_RATES
from ftw.shop.utils import is_email_valid
from plone.theme.interfaces import IDefaultPloneLayer
from z3c.form import validator
from z3c.form.interfaces import IRadioWidget
from z3c.form.validator import SimpleFieldValidator
from zope import schema
from zope.interface import Interface
from zope.interface import Invalid


class IShopRoot(Interface):
    """DEPRECATED!
    Former marker interface for shop root folder.
    """


class IBuyable(Interface):
    """Marker interface for marking items as buyable.
    """


class IShopItem(Interface):
    """A simple shop item
    """


class IShopCategory(Interface):
    """A category for shop items
    """


class IVariationConfig(Interface):
    """A component which provides variation configurations.
    """

    def getVariations(self):
        """Returns the variations stored on a ShopItem.
        """


class IShopOrder(Interface):
    """A shop order
    """


class ICartItem(Interface):
    """A shop item that has been added to the cart and is part of an order
    """


class IMailHostAdapter(Interface):
    """Adapter to abstract Plone 3 and Plone 4 MailHosts
    """

    def send(self, msg_body, mto, mfrom=None, mbcc=None, subject=None,
             encode=None, immediate=False, charset=None, msg_type=None):
        """Abstract sending mail with Plone 3 and Plone 4.
        """


# ------------ Storage related interfaces -------------

class IOrderStorage(Interface):
    """A local utility which stores the shop orders
    """

    def createOrder(self):
        pass

    def getOrder(self, order_id):
        pass

    def flush(self):
        pass



# ------------ Checkout Wizard related interfaces -------------

class IWizardStep(Interface):
    """A wizard step
    """


class IWizardStepGroup(Interface):
    """A wizard step group
    """


class IPaymentProcessor(Interface):
    """A payment processor
    """


class IContactInformationStepGroup(IWizardStepGroup):
    """A wizard step group gathering contact information about the customer
    """


class IContactInformationStep(IWizardStep):
    """A wizard step gathering contact information about the customer
    """


class IShippingAddressStepGroup(IWizardStepGroup):
    """A wizard step group gathering contact information about the
    preferred shipping address.
    """


class IShippingAddressStep(IWizardStep):
    """A wizard step gathering contact information about the preferred
    shipping address.
    """


class IPaymentProcessorChoiceStep(IWizardStep):
    """A wizard step asking to choose a payment processor
    """


class IDefaultPaymentProcessorChoice(Interface):
    """Schema defining a customer choice from enabled payment processors
    """
    payment_processor = schema.Choice(
            title=_(u"label_payment_processor", default="Payment Processor"),
            vocabulary="ftw.shop.enabled_payment_processors",
            required=True)


class IPaymentProcessorWidget(IRadioWidget):
    """Radio widget with fancy labels.
    """


class IPaymentProcessorStepGroup(IWizardStepGroup):
    """A wizard step group gathering contact information about the payment
    processor
    """


class IOrderReviewStep(IWizardStep):
    """A wizard step asking the customer to review the order and payment
    details and confirm the order.
    """


class IOrderReviewStepGroup(IWizardStepGroup):
    """A wizard step group for the order review step(s)
    """

# ------------------ Configuration Schemata -----------------------------

class IShopConfiguration(Interface):
    """This interface defines the ftw.shop configlet.
    """

    shop_name = schema.TextLine(
            title=_(u"label_shop_name", default=u"Enter the shop name"),
            required=True,
            default=u"Webshop")

    shop_email = schema.TextLine(
            title=_(u"label_shop_email", default=u"Sender e-Mail Address"),
            required=True,
            default=u"webshop@example.org")

    mail_bcc = schema.TextLine(
            title=_(u"label_mail_bcc", default=u"BCC Address"),
            required=False,
            default=u"")

    mail_subject = schema.TextLine(
            title=_(u"label_mail_subject",
                    default=u"Subject"),
            description=_(u"help_mail_subject",
                    default=u"The subject for order mails."),
            required=False,
            default=u"")

    phone_number = schema.TextLine(
        title=_(u'label_phone_number',
                default=u'Phone number'),
        description=_(u'help_phone_number',
                      default=u'Will be displayed in the customer confirmation mail'
                      ' as contact information when configured.'),
        required=False,
        default=u'')

    always_notify_shop_owner = schema.Bool(
            title=_(u"label_always_notify_shop_owner",
                    default=u"Always notify shop owner about orders"),
            description=_(u"help_always_notify_shop_owner",
                    default=u"Also send eMail to owner if supplier(s) have "
                             "been notified"),
            required=False)

    payment_processor_step_group = schema.Choice(
            title=_(u"label_payment_processor_step_group",
                    default="Payment Processor Step Group"),
            vocabulary="ftw.shop.payment_processor_step_groups",
            default=u"ftw.shop.DefaultPaymentProcessorStepGroup",
            required=True)

    enabled_payment_processors = schema.List(
            title=_(u"label_enabled_payment_processors",
                    default="Enabled Payment Processors"),
            required=False,
            default=[u'ftw.shop.InvoicePaymentProcessor'],
            value_type=schema.Choice(
                            vocabulary="ftw.shop.payment_processors"),
                            )

    order_storage = schema.Choice(
            title=_(u"label_order_storage",
                    default="Order Storage method"),
            required=True,
            default=u'ftw.shop.BTreeOrderStorage',
            vocabulary="ftw.shop.order_storage_vocabulary")

    contact_info_step_group = schema.Choice(
            title=_(u"label_contact_info_step_group",
                    default="Contact Information Step Group"),
            vocabulary="ftw.shop.contact_info_step_groups",
            default=u"ftw.shop.DefaultContactInformationStepGroup",
            required=True)

    shipping_address_step_group = schema.Choice(
            title=_(u"label_shipping_address_step_group",
                    default="Shipping Address Step Group"),
            vocabulary="ftw.shop.shipping_address_step_groups",
            default=u"ftw.shop.DefaultShippingAddressStepGroup",
            required=True)

    order_review_step_group = schema.Choice(
            title=_(u"label_order_review_step_group",
                    default="Order Review Step Group"),
            vocabulary="ftw.shop.order_review_step_groups",
            default=u"ftw.shop.DefaultOrderReviewStepGroup",
            required=True)

    status_set = schema.Choice(
            title=_(u"label_status_set",
                    default="Status Set"),
            vocabulary="ftw.shop.status_sets_vocabulary",
            default=u'ftw.shop.DefaultStatusSet',
            required=True)

    show_status_column = schema.Bool(
            title=_(u"label_show_status_column",
                    default=u"Show status column in order manager"),
            description=_(u"help_show_status_column",
                    default=u""),
            default=True,
            required=False)

    vat_enabled = schema.Bool(
            title=_(u"label_vat_enabled",
                    default=u"Enable VAT support"),
            description=_(u"help_vat_enabled",
                    default=u""),
            default=True,
            required=False)

    vat_number = schema.TextLine(
            title=_(u"label_vat_number",
                    default=u"VAT number"),
            description=_(u"help_vat_number",
                    default=u""),
            default=u'',
            required=False)

    vat_rates = schema.List(
            title=_(u"label_vat_rates",
                    default=u"VAT rates"),
            description=_(u"help_vat_rates",
                    default=u""),
            value_type=schema.Decimal(),
            default=DEFAULT_VAT_RATES,
            required=False)


class IDefaultContactInformation(Interface):
    """Schema defining a common contact address form
    """
    title= schema.TextLine(
            title=_(u'label_title', default=u'Title'),
            required=True)

    firstname = schema.TextLine(
            title=_(u'label_firstname', default=u'First Name'),
            required=True)

    lastname = schema.TextLine(
            title=_(u'label_lastname', default=u'Last Name'),
            required=True)

    email = schema.TextLine(
            title=_(u'label_email', default=u'Email'),
            required=True)

    company = schema.TextLine(
            title=_(u'label_company', default=u'Company'),
            required=False)

    street1 = schema.TextLine(
            title=_(u'label_street', default=u'Street/No.'),
            required=True)

    street2 = schema.TextLine(
            title=_(u'label_street2', default=u'Address 2'),
            required=False)

    phone = schema.TextLine(
            title=_(u'label_phone_number', default=u'Phone number'),
            required=True)

    zipcode = schema.TextLine(
            title=_(u'label_zipcode', default=u'Zip Code'),
            required=True)

    city = schema.TextLine(
            title=_(u'label_city', default=u'City'),
            required=True)

    country = schema.TextLine(
            title=_(u'label_country', default=u'Country'),
            required=True)


class IOrderReviewSchema(Interface):
    """Fields for the order review step.
    """

    comments = schema.Text(
            title=_(u'label_comments', default=u'Comments'),
            required=False)


class EmailAddressValidator(SimpleFieldValidator):

    def validate(self, value):
        super(EmailAddressValidator, self).validate(value)
        if not value or not is_email_valid(value):
            raise Invalid(_(u'This email address is invalid.'))


validator.WidgetValidatorDiscriminators(
    EmailAddressValidator,
    field=IDefaultContactInformation['email'])


class IShippingAddress(Interface):
    """Schema defining a form for entering a shipping address
    """

    used = schema.Bool(
            title=_(u'label_used', default=u'Different from invoice address'),
            default=False,
            required=False)

class IStatusSet(Interface):
    """Interface defining a set of statuses that an order can
    be in.
    """


class ISupplier(Interface):
    """A supplier
    """


class IFtwShopSpecific(IDefaultPloneLayer):
    """Marker interface for a zope 3 browser layer.
    """


class IShoppingCart(Interface):
    """ShoppingCart adapter
    """



from decimal import Decimal
from decimal import InvalidOperation
from ftw.shop import shopMessageFactory as _
from ftw.shop.interfaces import IVariationConfig, IShopItem
from persistent.mapping import PersistentMapping
from persistent.list import PersistentList
from zope.annotation.interfaces import IAnnotations
from zope.component import adapts
from zope.component import getMultiAdapter
from zope.globalrequest import getRequest
from zope.i18n import translate
from zope.interface import implements
import itertools


try:
    dummy = type(all)
except NameError:
    # Python 2.4
    def all(iterable):
        for element in iterable:
            if not element:
                return False
        return True


class VariationConfig(object):
    """An Adapter for storing variation configurations on ShopItems
    """

    implements(IVariationConfig)
    adapts(IShopItem)

    def __init__(self, context):
        self.context = context
        self.annotations = IAnnotations(self.context)

    def hasVariations(self):
        """Determines if the item has variations or not
        """
        field = self.context.Schema().getField('variation1_attribute')
        return field.get(self.context) not in (None, '')

    def getItemUID(self):
        """Returns the ShopItem's UID
        """
        return self.context.UID()

    def variation_code(self, var1choice=None, var2choice=None):
        """
        """
        if var1choice is None:
            return ''
        if var2choice is None:
            return 'var-%s' % var1choice
        return 'var-%s-%s' % (var1choice, var2choice)

    def sku_code(self, var1choice=None, var2choice=None):
        """Returns the sku code for the given variation combination.
        """
        vcode = self.variation_code(var1choice, var2choice)
        if not vcode:
            return self.context.getField('skuCode').get(self.context)
        return self.getVariationDict().get(vcode).get('skuCode')

    def getVariationDict(self):
        # Returns a nested dict with the variation config for the item

        return self.make_recursive_serializable(
            self.annotations.get('variations', PersistentMapping()))

    def updateVariationConfig(self, data):
        # Updates the stored variation config with changes

        if 'variations' not in self.annotations.keys():
            self.annotations['variations'] = PersistentMapping()

        self.annotations['variations'].update(
            self.make_recursive_persistent(data))

    def make_recursive_persistent(self, data):
        # Convert the data to a mapping for annotation-storage

        def persist(data):
            if isinstance(data, dict):
                data = PersistentMapping(data)

                for key, value in data.items():
                    data[key] = persist(value)

            elif isinstance(data, list):
                return PersistentList(map(persist, data))

            else:
                # Usually we got basestrings, or integer here, so do nothing.
                pass

            return data

        return persist(data)

    def make_recursive_serializable(self, data):
        # Convert the data to a mapping for annotation-storage

        def serialize(data):
            if isinstance(data, PersistentMapping):
                data = dict(data)

                for key, value in data.items():
                    data[key] = serialize(value)

            elif isinstance(data, PersistentList):
                return map(serialize, data)

            else:
                # Usually we got basestrings, or integer here, so do nothing.
                pass

            return data

        return serialize(data)

    def getVariation1Values(self):
        """Returns the values for the top level variation,
        e.g. ['Red', 'Green', 'Blue']
        """
        values = self.context.getField('variation1_values').get(self.context)
        if values:
            return values
        else:
            return []

    def getVariation2Values(self):
        """Returns the values for the second level variation,
        e.g. ['S', 'M', 'L', 'XL']
        """
        values = self.context.getField('variation2_values').get(self.context)
        if values:
            return values
        else:
            return []

    def getVariationAttributes(self):
        """Returns a list of the two variation attributes,
        e.g. ['Color', 'Size']
        """
        variation_attributes = []
        for name in ['variation1_attribute', 'variation2_attribute']:
            field = self.context.Schema().getField(name)
            if field.get(self.context) not in (None, ''):
                variation_attributes.append(field.get(self.context))
        return variation_attributes

    def getVariationData(self, var1_idx, var2_idx, field):
        """Returns the data for one specific variation instance's field
        """
        variation_dict = self.getVariationDict()

        if var2_idx is None:
            # We only have one variation
            variation_code = "var-%s" % var1_idx
        else:
            # We have two levels of variation
            variation_code = "var-%s-%s" % (var1_idx, var2_idx)
        var_data = variation_dict.get(variation_code, None)
        if var_data is not None and field in var_data.keys():
            if not var_data[field] == "":
                return var_data[field]

        # Return a default value appropriate for the field type
        if field == 'active':
            return True
        elif field == 'price':
            return Decimal("%s.%02d" % self.context.price)
        elif field == 'skuCode':
            return self.context.skuCode
        elif field == 'description':
            return ''
        elif field == 'hasUniqueSKU':
            return False
        else:
            return None

    def getPrettyName(self, variation_code):
        """Returns the human facing name for a variation,
        e.g. 'Green-XXL'
        """
        if len(self.getVariationAttributes()) == 1:
            for i, var1_value in enumerate(self.getVariation1Values()):
                vcode = "var-%s" % i
                if vcode == variation_code:
                    return var1_value
        else:
            for i, var1_value in enumerate(self.getVariation1Values()):
                for j, var2_value in enumerate(self.getVariation2Values()):
                    vcode = "var-%s-%s" % (i, j)
                    if vcode == variation_code:
                        return "%s-%s" % (var1_value, var2_value)
        return None


    def isValid(self):
        # Since skuCode uniqueness isn't required any more,
        # variation configs are always assumed to be valid
        return True


    def allPricesZero(self):
        var_dict = self.getVariationDict()
        try:
            prices = [Decimal(var_dict[k].get('price', '0.0'))
                            for k in var_dict]
            return all([p == Decimal('0.0') for p in prices])
        except InvalidOperation:
            return False

    def purge_dict(self):
        self.annotations['variations'] = PersistentMapping()

    def reduce_level(self):
        if len(self.getVariationAttributes()) == 2:
            # Reduce from 2 to 1 variation
            var_dict = self.getVariationDict()
            new_var_dict = PersistentMapping()
            vcodes = sorted(var_dict.keys())
            for i, vcode in enumerate(vcodes):
                new_vcode = 'var-%s' % i
                new_var_dict[new_vcode] = var_dict[vcode]

            self.annotations['variations'] = new_var_dict
            new_values = []

            combinations = itertools.product(self.getVariation1Values(), self.getVariation2Values())
            for pair in combinations:
                val = '-'.join(pair)
                new_values.append(val)

            new_attr = "%s/%s" % (self.getVariationAttributes()[0],
                                  self.getVariationAttributes()[1])

            self.context.getField('variation2_values').set(self.context, [])
            self.context.getField('variation2_attribute').set(self.context, None)

            self.context.getField('variation1_values').set(self.context, new_values)
            self.context.getField('variation1_attribute').set(self.context, new_attr)

        elif len(self.getVariationAttributes()) == 1:
            # Reduce from 1 to 0 variations
            self.context.getField('variation1_values').set(self.context, [])
            self.context.getField('variation1_attribute').set(self.context, None)

            self.context.getField('variation2_values').set(self.context, [])
            self.context.getField('variation2_attribute').set(self.context, None)

            self.annotations['variations'] = PersistentMapping()


    def remove_level(self):
        if len(self.getVariationAttributes()) == 2:
            self.context.getField('variation2_values').set(self.context, [])
            self.context.getField('variation2_attribute').set(self.context, None)
            self.purge_dict()
        elif len(self.getVariationAttributes()) == 1:
            self.context.getField('variation1_values').set(self.context, [])
            self.context.getField('variation1_attribute').set(self.context, None)
            self.context.getField('variation2_values').set(self.context, [])
            self.context.getField('variation2_attribute').set(self.context, None)
            self.purge_dict()


    def add_level(self):
        fields = ['active', 'skuCode', 'price', 'description']
        request = getRequest()
        pps = getMultiAdapter((self.context, request), name='plone_portal_state')
        language = pps.language()
        if len(self.getVariationAttributes()) == 1:
            new_value_1 = translate(_('label_new_value_1', default=u'New value 1'), domain='ftw.shop', context=self.context, target_language=language)
            new_value_2 = translate(_('label_new_value_2', default=u'New value 2'), domain='ftw.shop', context=self.context, target_language=language)
            new_attr = translate(_('label_new_attr', default=u'New attribute'), domain='ftw.shop', context=self.context, target_language=language)
            self.context.getField('variation2_values').set(self.context, [new_value_1, new_value_2])
            self.context.getField('variation2_attribute').set(self.context, new_attr)

            # Initialize var data for newly added level with default values
            for i in range(len(self.getVariation1Values())):
                for j in range(len(self.getVariation2Values())):
                    vardata = {}
                    for f in fields:
                        vcode = "var-%s-%s" % (i, j)
                        data = self.getVariationData(i, j, f)
                        vardata[f] = data
                        partial_vardict = {vcode: vardata}
                        self.updateVariationConfig(partial_vardict)

        elif len(self.getVariationAttributes()) == 0:
            new_value_1 = translate(_('label_new_value_1', default=u'New value 1'), domain='ftw.shop', context=self.context, target_language=language)
            new_value_2 = translate(_('label_new_value_2', default=u'New value 2'), domain='ftw.shop', context=self.context, target_language=language)
            new_attr = translate(_('label_new_attr', default=u'New attribute'), domain='ftw.shop', context=self.context, target_language=language)

            self.context.getField('variation1_values').set(self.context, [new_value_1, new_value_2])
            self.context.getField('variation1_attribute').set(self.context, new_attr)

            # Initialize var data for newly added level with default values
            for i in range(len(self.getVariation1Values())):
                vardata = {}
                for f in fields:
                    vcode = "var-%s" % (i)
                    data = self.getVariationData(i, None, f)
                    vardata[f] = data
                    partial_vardict = {vcode: vardata}
                    self.updateVariationConfig(partial_vardict)



from ftw.shop import shopMessageFactory as _
from ftw.shop.content.shopitem import selectable_dimensions
from ftw.shop.interfaces import IContactInformationStepGroup
from ftw.shop.interfaces import IOrderReviewStepGroup
from ftw.shop.interfaces import IOrderStorage
from ftw.shop.interfaces import IPaymentProcessor
from ftw.shop.interfaces import IPaymentProcessorStepGroup
from ftw.shop.interfaces import IShippingAddressStepGroup
from ftw.shop.interfaces import IShopConfiguration
from ftw.shop.interfaces import IStatusSet
from functools import partial
from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from zope.component import getAdapters, getUtility, getUtilitiesFor
from zope.component.hooks import getSite
from zope.i18n import translate
from zope.interface import directlyProvides
from zope.schema import vocabulary
from zope.schema.interfaces import IVocabularyFactory
from zope.schema.vocabulary import SimpleTerm


def create_terms_from_adapters(adapters):
    """Returns a list of terms to be used to create a vocabulary,
    including a nice human readable title if defined for the adapter
    """
    adapter_names = []
    adapter_titles = []
    items = []
    for name, adapter in adapters:
        adapter_names.append(unicode(name))
        title = getattr(adapter, 'title', name)
        adapter_titles.append(title)

    for i in range(0, len(adapter_names)):
        items.append(tuple([adapter_names[i], adapter_titles[i]]))

    items.sort()
    terms = [SimpleTerm(value=pair[0],
                        token=pair[0],
                        title=pair[1]) for pair in items]
    return terms


def ContactInfoStepGroups(context):
    """Returns a vocabulary of the registered StepGroups for the
    contact information StepGroup
    """
    request = getSite().REQUEST
    # context is the portal config options, whose context is the portal
    contact_info_step_groups = getAdapters((context, request, context),
                                           IContactInformationStepGroup)
    terms = create_terms_from_adapters(contact_info_step_groups)

    directlyProvides(ContactInfoStepGroups, IVocabularyFactory)
    return vocabulary.SimpleVocabulary(terms)


def ShippingAddressStepGroups(context):
    """Returns a vocabulary of the registered StepGroups for the
    shipping address StepGroup
    """
    # context is the portal config options, whose context is the portal
    request = getSite().REQUEST
    shipping_address_step_groups = getAdapters((context, request, context),
                                               IShippingAddressStepGroup)
    terms = create_terms_from_adapters(shipping_address_step_groups)

    directlyProvides(ShippingAddressStepGroups, IVocabularyFactory)
    return vocabulary.SimpleVocabulary(terms)


def OrderReviewStepGroups(context):
    """Returns a vocabulary of the registered StepGroups for the
    order review StepGroup
    """
    # context is the portal config options, whose context is the portal
    request = getSite().REQUEST
    order_review_step_groups = getAdapters((context, request, context),
                                           IOrderReviewStepGroup)
    terms = create_terms_from_adapters(order_review_step_groups)

    directlyProvides(OrderReviewStepGroups, IVocabularyFactory)
    return vocabulary.SimpleVocabulary(terms)


def PaymentProcessorStepGroups(context):
    """Returns a vocabulary of the registered StepGroups for the
    payment processor StepGroup
    """
    # context is the portal config options, whose context is the portal
    request = getSite().REQUEST
    payment_processor_step_groups = getAdapters((context, request, context),
                                                IPaymentProcessorStepGroup)
    terms = create_terms_from_adapters(payment_processor_step_groups)

    directlyProvides(PaymentProcessorStepGroups, IVocabularyFactory)

    return vocabulary.SimpleVocabulary(terms)


def PaymentProcessors(context):
    """Returns a vocabulary of the registered PaymentProcessors
    """
    # context is the portal config options, whose context is the portal
    request = getSite().REQUEST
    payment_processors = getAdapters((context, request, context),
                                     IPaymentProcessor)
    terms = create_terms_from_adapters(payment_processors)

    directlyProvides(PaymentProcessors, IVocabularyFactory)

    return vocabulary.SimpleVocabulary(terms)


def EnabledPaymentProcessors(context):
    """Returns a vocabulary of the PaymentProcessors that have been enabled
    in the shop configuration.
    """
    # context is the portal config options, whose context is the portal
    registry = getUtility(IRegistry)
    shop_config = registry.forInterface(IShopConfiguration)
    request = getSite().REQUEST
    payment_processors = getAdapters((context, request, context),
                                     IPaymentProcessor)
    terms = create_terms_from_adapters(payment_processors)
    for term in terms:
        if term.value not in shop_config.enabled_payment_processors:
            terms.remove(term)

    directlyProvides(EnabledPaymentProcessors, IVocabularyFactory)
    return vocabulary.SimpleVocabulary(terms)


def OrderStorageVocabulary(context):
    """Returns a vocabulary of the currently registered utilities
    that implement IOrderStorage.
    """
    # context is the portal config options, whose context is the portal
    order_storages = getUtilitiesFor(IOrderStorage)
    terms = create_terms_from_adapters(order_storages)

    directlyProvides(OrderStorageVocabulary, IVocabularyFactory)
    return vocabulary.SimpleVocabulary(terms)


def StatusSetsVocabulary(context):
    """Returns a vocabulary of the currently registered utilities
    that implement IStatusSet.
    """
    # context is the portal config options, whose context is the portal
    status_sets = getAdapters((context,), IStatusSet)
    terms = create_terms_from_adapters(status_sets)

    directlyProvides(StatusSetsVocabulary, IVocabularyFactory)
    return vocabulary.SimpleVocabulary(terms)


def VATRatesVocabulary(context):
    """Returns a vocabulary of the VAT rates that are available
    to choose from.
    """
    # context is the portal config options, whose context is the portal
    registry = getUtility(IRegistry)
    shop_config = registry.forInterface(IShopConfiguration)
    vat_rates = shop_config.vat_rates

    terms = [SimpleTerm(value=str(rate), token=rate, title=str(rate))
             for rate in vat_rates]

    directlyProvides(VATRatesVocabulary, IVocabularyFactory)
    return vocabulary.SimpleVocabulary(terms)


def SuppliersVocabulary(context):
    """Returns a vocabulary of the available suppliers
    """
    # context is the portal config options, whose context is the portal
    catalog = getToolByName(context, 'portal_catalog')
    suppliers = catalog(portal_type="Supplier")
    items = [(brain.UID, brain.Title) for brain in suppliers]
    terms = [SimpleTerm(value=pair[0],
                        token=pair[0],
                        title=pair[1]) for pair in items]
    terms.insert(0, SimpleTerm(value='',
                               token='none',
                               title=_("label_supplier_from_parent",
                                       default="Supplier from parent category")))

    directlyProvides(SuppliersVocabulary, IVocabularyFactory)
    return vocabulary.SimpleVocabulary(terms)


def SelectableDimensionsVocabulary(context):
    """Returns a vocabulary of the available suppliers
    """
    transl = partial(translate, context=context.REQUEST)

    terms = []
    for key, value in selectable_dimensions.iteritems():
        title = transl(value['label'])
        if key != 'no_dimensions':
            form = transl(_(
                u'dimensions_format',
                default=u'%(label)s (%(dimension)s) - price in %(price)s',
            ))
            title = form % {
                'label': title,
                'dimension': transl(value['dimension_unit']),
                'price': transl(value['price_unit'] if 'price_unit' in value else translate(value['dimension_unit']))
            }

        terms.append(SimpleTerm(value=key, token=key, title=title))

    terms = sorted(terms, key=lambda term: term.title)

    directlyProvides(SelectableDimensionsVocabulary, IVocabularyFactory)
    return vocabulary.SimpleVocabulary(terms)



from ftw.builder import Builder
from ftw.builder import create
from ftw.shop.tests.pages import shopitem
from ftw.testbrowser import browser


def visit_edit_categories(item=None):
    if item is None:
        item = create(Builder('shop item'))
    browser.visit(item)
    shopitem.edit_categories_link().click()


def review_category_tree():
    category_tree = []
    listing_tables = browser.css('.categoryListing')[1:]
    for table in listing_tables:
        category_branch = []
        for row in table.css('tr'):
            expand_toggle, checkbox, name, rank = row.css('td')
            item_title = name.text
            breadcrumb_title = name.css('a').first.attrib.get('title', '')
            category_branch.append([item_title, breadcrumb_title])

        category_tree.append(category_branch)
    return category_tree


from decimal import Decimal
import simplejson
from Acquisition import aq_inner
from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from plone.memoize import instance

from ftw.shop.config import CATEGORY_RELATIONSHIP
from ftw.shop.interfaces import IShopItem
from ftw.shop.interfaces import IVariationConfig


class CategoryView(BrowserView):
    """Default view for a category. Shows all contained items and categories.
    """

    __call__ = ViewPageTemplateFile('templates/category.pt')

    single_item_template = ViewPageTemplateFile('templates/listing/single_item.pt')
    one_variation_template = ViewPageTemplateFile('templates/listing/one_variation_compact.pt')
    two_variations_template = ViewPageTemplateFile('templates/listing/two_variations_compact.pt')


    def getItems(self):
        """Returns a list of ShopItems directly contained in this category
        """
        return [item for item in self.category_contents
                if IShopItem.providedBy(item)]

    def single_item(self, item):
        return self.single_item_template(item=item)

    def one_variation(self, item):
        return self.one_variation_template(item=item)

    def two_variations(self, item):
        return self.two_variations_template(item=item)

    def getItemDatas(self):
        """Returns a dictionary of an item's properties to be used in
        templates. If the item has variations, the variation config is
        also included.
        """
        results = []
        for item in self.getItems():
            assert(IShopItem.providedBy(item))
            varConf = IVariationConfig(item)

            has_variations = varConf.hasVariations()

            image = item.getField('image')
            if image and image.get_size(item):
                hasImage = True
                tag = image.tag(item, scale='mini')
            else:
                hasImage = False
                tag = None
            if has_variations:
                skuCode = None
                price = None
            else:
                varConf = None
                skuCode = item.Schema().getField('skuCode').get(item)
                price = item.Schema().getField('price').get(item)

            results.append(
                dict(
                    item = item,
                    title = item.Title(),
                    description = item.Description(),
                    url = item.absolute_url(),
                    hasImage = hasImage,
                    imageTag = tag,
                    variants = None,
                    skuCode = skuCode,
                    price = price,
                    showPrice = item.getField('showPrice').get(item),
                    unit=item.getField('unit').get(item),
                    uid = item.UID(),
                    varConf = varConf,
                    hasVariations = has_variations,
                    selectable_dimensions = item.getSelectableDimensions()))
        return results

    def getVarDictsJSON(self):
        """Returns a JSON serialized dict with UID:varDict pairs, where UID
        is the ShopItem's UID and varDict is the item's variation dict.
        This is being used for the compact category view where inactive
        item variations must not be buyable.
        """
        varDicts = {}
        items = self.getItemDatas()
        for item in items:
            uid = item['uid']
            varConf = item['varConf']
            if varConf is not None:
                varDicts[uid] = dict(varConf.getVariationDict())
            else:
                varDicts[uid] = {}

            # Convert Decimals to Strings for serialization
            varDict = varDicts[uid]
            for vcode in varDict.keys():
                i = varDict[vcode]
                for k in i.keys():
                    val = i[k]
                    if isinstance(val, Decimal):
                        val = str(val)
                        i[k] = val

        return simplejson.dumps(varDicts)

    @property
    @instance.memoize
    def categories(self):
        """ get a list with all categories belonging to this category.
        """

        results = []
        for item in self.category_contents:
            image = None
            tag = None
            if image:
                tag = image.tag(scale='tile')
            if item.portal_type == 'ShopCategory':
                results.append(dict(
                    title = item.Title(),
                    description = item.Description(),
                    url = item.absolute_url(),
                    image = tag,
                ))
        return results

    @property
    @instance.memoize
    def category_contents(self):
        """ get all items (shop items, categories) belonging to this category.
        """
        def filter_language(item):
            """Return ``True`` for items that have the same language as the
               context or items that have no language.
            """
            language = context.Language()
            if not item.Language():
                return True
            if item.Language()==language:
                return True
            return False

        context = aq_inner(self.context)
        mtool = getToolByName(context, 'portal_membership')

        contents = context.getBRefs(CATEGORY_RELATIONSHIP)

        contents = [item for item in contents
                    if item and mtool.checkPermission('View', item)
                    and filter_language(item)]

        # Sort alphabetically first
        contents.sort(key=lambda x: x.Title())

        # Then by ranking
        contents.sort(lambda x, y: cmp(x.getRankForCategory(context),
                        y.getRankForCategory(context)))
        return contents



# -*- coding: utf-8 -*-
from ftw.shop.browser.ordermanager import COLUMN_TITLES
from ftw.shop.config import CART_KEY
from ftw.shop.config import SESSION_ADDRESS_KEY
from ftw.shop.config import SESSION_SHIPPING_KEY
from ftw.shop.tests.base import FtwShopTestCase
from ftw.shop.tests.base import MOCK_CART
from ftw.shop.tests.base import MOCK_CUSTOMER
from ftw.shop.tests.base import MOCK_SHIPPING
from plone import api
from zope.component import  getMultiAdapter
import unittest


class TestOrderManager(FtwShopTestCase):

    def afterSetUp(self):
        super(TestOrderManager, self).afterSetUp()

    def test_download_csv(self):
        session = self.portal.REQUEST.SESSION
        # Create an order that can be used to test CSV generation
        navroot = api.portal.get_navigation_root(self.portal)
        order_manager = getMultiAdapter((navroot, self.portal.REQUEST),
                                   name=u'order_manager')
        session[SESSION_ADDRESS_KEY] = MOCK_CUSTOMER
        session[SESSION_SHIPPING_KEY] = MOCK_SHIPPING
        session[CART_KEY] = MOCK_CART
        session['order_confirmation'] = True
        session['payment_processor_choice'] = {'payment_processor': 'Test'}
        order_id = order_manager.addOrder()
        order = order_manager.getOrder(order_id)
        order_date = order.date
        order_number = order.title

        # Call the OrderManager view to initialize om.order_results
        self.login()
        order_manager()

        csv = order_manager.download_csv()
        csv_header = csv.split('\r\n')[0]
        csv_data = csv.split('\r\n')[1]

        expected_header_ids = [
            'order_id',
            'title',
            'status',
            'total',
            'date',
            'customer_title',
            'customer_firstname',
            'customer_lastname',
            'customer_company',
            'customer_email',
            'customer_street1',
            'customer_street2',
            'customer_phone',
            'customer_zipcode',
            'customer_city',
            'customer_shipping_address',
            'customer_country',
            'customer_comments',
            'shipping_city',
            'shipping_company',
            'shipping_firstname',
            'shipping_lastname',
            'shipping_street1',
            'shipping_street2',
            'shipping_title',
            'shipping_zipcode',
            'vat_amount',
            'sku_code',
            'quantity',
            'dimensions',
            'title',
            'price',
            'item_total',
            'supplier_name',
            'supplier_email']
        expected_csv_header = ';'.join([COLUMN_TITLES[h] for h in expected_header_ids])

        expected_csv_data = ';'.join([
            '1',
            order_number,
            '1',
            str(order.getTotal()),
            str(order_date),
            MOCK_CUSTOMER['title'],
            MOCK_CUSTOMER['firstname'],
            MOCK_CUSTOMER['lastname'],
            MOCK_CUSTOMER['company'],
            MOCK_CUSTOMER['email'],
            MOCK_CUSTOMER['street1'],
            '',
            MOCK_CUSTOMER['phone'],
            MOCK_CUSTOMER['zipcode'],
            MOCK_CUSTOMER['city'],
            '',
            MOCK_CUSTOMER['country'],
            '',
            MOCK_SHIPPING['city'],
            MOCK_SHIPPING['company'],
            MOCK_SHIPPING['firstname'],
            MOCK_SHIPPING['lastname'],

            MOCK_SHIPPING['street1'],
            '',
            MOCK_SHIPPING['title'],
            MOCK_SHIPPING['zipcode'],

            MOCK_CART['some-uid']['vat_amount'],
            MOCK_CART['some-uid']['skucode'],
            str(MOCK_CART.values()[0]['quantity']),
            '2 Weight (g)',
            MOCK_CART.values()[0]['title'],
            str(MOCK_CART.values()[0]['price']),
            str(MOCK_CART.values()[0]['total']),
            MOCK_CART.values()[0]['supplier_name'],
            MOCK_CART.values()[0]['supplier_email'],])

        self.assertEquals(csv_header, expected_csv_header)
        self.assertEquals(csv_data, expected_csv_data)


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestOrderManager))
    return suite
